# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import (
    INPUT_CHANNEL, INT32_ELEMENT, OUTPUT_CHANNEL, ChannelMetaData, Epochstamp,
    Schema, TimeId, Timestamp)


def test_metadata_general_functionality():
    es = Epochstamp(1356441936, 789333123456789123)
    ts = Timestamp(es, TimeId(987654321))
    meta = ChannelMetaData('abc', ts)

    assert meta.getSource() == 'abc'
    assert meta.getTimestamp().getTid() == 987654321
    assert meta.getTimestamp().getSeconds() == 1356441936
    assert meta.getTimestamp().getFractionalSeconds() == 789333123456789123
    assert meta.getTimestamp().toFormattedString() == ts.toFormattedString()
    # Equality operator is bound
    assert meta.getTimestamp() == ts

    meta.setSource('xyz')
    assert meta.getSource() != 'abc'
    assert meta.getSource() == 'xyz'

    ts1 = Timestamp()
    meta.setTimestamp(ts1)
    assert meta.getTimestamp().getTid() == 0
    assert meta.getTimestamp().toFormattedString() != ts.toFormattedString()
    assert meta.getTimestamp().toFormattedString() == ts1.toFormattedString()


def test_channel_schema_declarations():
    pipeSchema = Schema()
    INT32_ELEMENT(pipeSchema).key("int32").readOnly().commit()

    s = Schema()
    (
        OUTPUT_CHANNEL(s)
        .key('validkey')
        .displayedName('ValidKey')
        .description('Valid key description')
        .dataSchema(pipeSchema)
        .commit(),

        INPUT_CHANNEL(s)
        .key('input')
        .displayedName('Input key')
        .description('Input key description')
        .commit(),
    )

    assert s.has("validkey") and s.has("input") is True
    assert s.getDisplayedName('validkey') == 'ValidKey'
    assert s.getDisplayedName('input') == 'Input key'
    assert s.getDescription('validkey') == 'Valid key description'
    assert s.getDescription('input') == 'Input key description'
    assert s.has("validkey.schema.int32") is True



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
'''
Created on Oct 17, 2012

@author: Sergey Esenov <serguei.essenov@xfel.eu>
'''

import gc
import unittest
import weakref

import numpy as np

from karabo.bound import Hash, Types
from karabo.testing.utils import compare_ndarray_data_ptrs


class Hash_TestCase(unittest.TestCase):

    def test_intUnboxingEdgeCases(self):
        values_and_types = {
            -(2 ** 31): (Types.INT32, Types.VECTOR_INT32),
            2 ** 31 - 1: (Types.INT32, Types.VECTOR_INT32),
            2 ** 32 - 1: (Types.UINT32, Types.VECTOR_UINT32),
            2 ** 32: (Types.INT64, Types.VECTOR_INT64),
            -(2 ** 63): (Types.INT64, Types.VECTOR_INT64),
            2 ** 63 - 1: (Types.INT64, Types.VECTOR_INT64),
            2 ** 64 - 1: (Types.UINT64, Types.VECTOR_UINT64),
        }
        # Why not OverflowError below? If I try so, my traceback is
        # OverflowError: int too big to convert
        #
        # The above exception was the direct cause of the following exception:
        #
        # Traceback (most recent call last):
        #   File "<snip>/tests/hash_test.py", line 397, in test_<snip>
        #  h = Hash("b", 2 ** 64)
        # SystemError: <Boost.Python.function object at 0x242c6c0> returned a
        #              result with an error set
        with self.assertRaises(RuntimeError):
            h = Hash("b", 2 ** 64)
        with self.assertRaises(RuntimeError):
            h = Hash("vb", [2 ** 64])
        for value, type_ in values_and_types.items():
            msg = f"Failed to unbox {value}: "
            try:
                h = Hash("a", value)
                self.assertEqual(h.getType("a"), type_[0], msg=msg)
                self.assertEqual(h["a"], value, msg=msg)
            except Exception as e:
                self.fail(msg + repr(e))
            # Now for vectors:
            # first most broad int type first
            msg = f"Failed to unbox {value} in list: "
            try:
                h = Hash("a", [value, 0])
                self.assertEqual(h.getType("a"), type_[1], msg=msg)
                self.assertEqual(h["a"], [value, 0], msg=msg + " - first")
            except Exception as e:
                self.fail(msg + repr(e))
            # then most broad int type last
            try:
                h = Hash("a", [0, value])
                self.assertEqual(h.getType("a"), type_[1], msg=msg)
                self.assertEqual(h["a"], [0, value], msg=msg + " - last")
            except Exception as e:
                self.fail(msg + repr(e))

    def test_numpy_arrays(self):
        arr = np.ones((100,), dtype=np.uint32)
        arr_weak = weakref.ref(arr)

        h = Hash("a", arr)
        assert compare_ndarray_data_ptrs(h["a"], arr)  # no copy

        # Make sure references are cleaned up appropriately
        del arr
        assert arr_weak() is not None  # h is still holding a reference
        del h
        assert arr_weak() is None

    def test_find(self):
        try:
            h = Hash("a.b.c1.d", 1)
            node = h.find("a.b.c1.d")
            self.assertIsNotNone(node,
                                 "The 'node' object should be not 'None'")
            self.assertEqual(node.getKey(), "d",
                             'Bad key returned by "getKey" method')
            self.assertEqual(node.getValue(), 1, 'Should return 1')
            self.assertIsNone(h.find("a.b.c1.f"),
                              "The resulting object should be 'None'")
        except Exception as e:
            self.fail("test_find exception group 1: " + str(e))

        try:
            h = Hash("a.b.c", "1")
            node = h.find("a.b.c")
            self.assertIsNotNone(node, "The 'node' should be not 'None")
            node.setValue(2)
            self.assertEqual(h["a.b.c"], 2)
            node = h.find("a.b.c", '/')
            self.assertIsNone(node)
        except Exception as e:
            self.fail("test_find exception group 2: " + str(e))

    def test_iteration(self):
        h = Hash("should", 1, "be", 2, "iterated", 3, "in", 4, "correct", 5,
                 "order", 6)
        try:
            insertionOrder = list()
            for k in h:
                insertionOrder.append(str(k))
            self.assertEqual(insertionOrder,
                             ["should", "be", "iterated", "in", "correct",
                              "order"])

        except Exception as e:
            self.fail("test_iteration exception group 1: " + str(e))

        try:
            alphaNumericOrder = list()
            for k in h:
                alphaNumericOrder.append(k.getKey())
            alphaNumericOrder.sort()
            self.assertEqual(alphaNumericOrder,
                             ["be", "correct", "in", "iterated", "order",
                              "should"])
        except Exception as e:
            self.fail("test_iteration exception group 2: " + str(e))

        h.set("be", "2")  # Has no effect on order

        try:
            insertionOrder = list()
            for k in h:
                insertionOrder.append(str(k.getKey()))
            self.assertEqual(insertionOrder,
                             ["should", "be", "iterated", "in", "correct",
                              "order"])
        except Exception as e:
            self.fail("test_iteration exception group 3: " + str(e))

        try:
            alphaNumericOrder = list()
            for k in h:
                alphaNumericOrder.append(str(k.getKey()))
            alphaNumericOrder.sort()
            self.assertEqual(alphaNumericOrder,
                             ["be", "correct", "in", "iterated", "order",
                              "should"])
        except Exception as e:
            self.fail("test_iteration exception group 4: " + str(e))

        h.erase("be")  # Remove
        h.set("be", "2")  # Must be last element in sequence now

        try:
            insertionOrder = list()
            for k in h:
                insertionOrder.append(str(k.getKey()))
            self.assertEqual(insertionOrder,
                             ["should", "iterated", "in", "correct", "order",
                              "be"])
        except Exception as e:
            self.fail("test_iteration exception group 5: " + str(e))

        try:
            alphaNumericOrder = list()
            for k in h:
                alphaNumericOrder.append(str(k.getKey()))
            alphaNumericOrder.sort()
            self.assertEqual(alphaNumericOrder,
                             ["be", "correct", "in", "iterated", "order",
                              "should"])
        except Exception as e:
            self.fail("test_iteration exception group 6: " + str(e))

        # getKeys(...) to ...
        #         "list" and sort it  ... like C++ set
        try:
            tmp = list()  # create empty set
            h.getKeys(tmp)  # fill set by keys
            tmp.sort()
            it = iter(tmp)
            self.assertEqual(str(next(it)), "be")
            self.assertEqual(str(next(it)), "correct")
            self.assertEqual(str(next(it)), "in")
            self.assertEqual(str(next(it)), "iterated")
            self.assertEqual(str(next(it)), "order")
            self.assertEqual(str(next(it)), "should")
        except Exception as e:
            self.fail("test_iteration exception group 7: " + str(e))

        # "list" ... like C++ vector
        try:
            tmp = list()  # create empty vector
            h.getKeys(tmp)  # fill vector by keys
            it = iter(tmp)
            self.assertEqual(str(next(it)), "should")
            self.assertEqual(str(next(it)), "iterated")
            self.assertEqual(str(next(it)), "in")
            self.assertEqual(str(next(it)), "correct")
            self.assertEqual(str(next(it)), "order")
            self.assertEqual(str(next(it)), "be")
        except Exception as e:
            self.fail("test_iteration exception group 8: " + str(e))

        try:
            h = Hash("b", "bla-la-la", "a.b.c", 1, "abc.2", 2.2222222,
                     "a.b.c1.d", "abc1d", "abc.1", 1.11111111)

            ll = []
            h.getKeys(ll)  # use top level keys
            self.assertEqual(len(ll), 3)
            i = iter(ll)  # "canonical" order: on every level insertion order
            self.assertEqual(str(next(i)), "b")
            self.assertEqual(str(next(i)), "a")
            self.assertEqual(str(next(i)), "abc")

            ll = []
            h.getPaths(ll)  # use full keys
            self.assertEqual(len(ll), 5)
            i = iter(ll)
            self.assertEqual(str(next(i)), "b")
            self.assertEqual(str(next(i)), "a.b.c")
            self.assertEqual(str(next(i)), "a.b.c1.d")
            self.assertEqual(str(next(i)), "abc.2")
            self.assertEqual(str(next(i)), "abc.1")
        except Exception as e:
            self.fail("test_iteration exception group 9: " + str(e))

        try:
            hash = Hash("a.b[0].c.d", "bla-la-la", "a.b[1].c.d", "bla-la-la")
            vec = hash['a.b']
            for h in vec:
                self.assertEqual(next(h.__iter__()).getValue()['d'],
                                 "bla-la-la")
        except Exception as e:
            self.fail("test_iteration exception group 10: " + str(e))

    def test_attributes(self):
        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", "someValue")
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')
        except Exception as e:
            self.fail("test_attributes exception group 1: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", "someValue")
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')

            h.setAttribute("a.b.a.b", "attr2", 42)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')
            self.assertEqual(h.getAttribute("a.b.a.b", "attr2"), 42,
                             'Should return 42')

            h.setAttribute("a.b.a.b", "attr2", 43)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')
            self.assertEqual(h.getAttribute("a.b.a.b", "attr2"), 43,
                             'Should return 42')

            h.setAttribute("a.b.a.b", "attr3", 1 + 2j)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr3"), 1 + 2j,
                             'Should return 1+2j')

            h.setAttribute("a.b.a.b", "attr1", True)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True,
                             'Should return "someValue"')
            self.assertEqual(h.getAttribute("a.b.a.b", "attr2"), 43,
                             'Should return 42')

            attrs = h.getAttributes("a.b.a.b")
            self.assertEqual(attrs.size(), 3)
            self.assertEqual(attrs.get("attr1"), True)
            self.assertEqual(attrs["attr1"], True)
            self.assertEqual(attrs.get("attr2"), 43)
            self.assertEqual(attrs["attr2"], 43)

            node = attrs.getNode("attr1")
            self.assertEqual(node.getType(), Types.BOOL)

            node = attrs.getNode("attr2")
            self.assertEqual(node.getType(), Types.INT32)

        except Exception as e:
            self.fail("test_attributes exception group 2: " + str(e))

        try:
            h = Hash("a.b.a.b", 1)
            h.setAttribute("a.b.a.b", "attr1", [1, 2, 3])
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1")[1], 2)
        except Exception as e:
            self.fail("test_attributes exception group 3: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", "someValue")
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')
        except Exception as e:
            self.fail("test_attributes exception group 4: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", "someValue")
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), "someValue",
                             'Should return "someValue"')
        except Exception as e:
            self.fail("test_attributes exception group 5: " + str(e))

    def test_attributes_get_copy(self):
        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", True)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)

            attrs = h.copyAttributes("a.b.a.b")
            attrs["attr1"] = False
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)
        except Exception as e:
            self.fail("test_attributes_get_copy exception group 1: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", True)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)

            attrs = h.getAttributes("a.b.a.b")
            attrs["attr1"] = False
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), False)
        except Exception as e:
            self.fail("test_attributes_get_copy exception group 2: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", True)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)
            node = h.getNode("a.b.a.b")
            attrs = node.copyAttributes()
            attrs["attr1"] = False
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)
        except Exception as e:
            self.fail("test_attributes_get_copy exception group 3: " + str(e))

        try:
            h = Hash("a.b.a.b", 42)
            h.setAttribute("a.b.a.b", "attr1", True)
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), True)
            node = h.getNode("a.b.a.b")
            attrs = node.getAttributes()
            attrs["attr1"] = False
            self.assertEqual(h.getAttribute("a.b.a.b", "attr1"), False)
        except Exception as e:
            self.fail("test_attributes_get_copy exception group 4: " + str(e))

    def test_dict(self):
        try:
            h = Hash("a", {"b": {"c": {"d": [1, 2, 3, 4, 5]}}})

            self.assertEqual(h["a.b.c.d"], [1, 2, 3, 4, 5])
            self.assertEqual(h["a"]["b"]["c"]["d"], [1, 2, 3, 4, 5])

            h.set('x', {'y': {'z': True}})

            self.assertEqual(h["x.y.z"], True)

        except Exception as e:
            self.fail("test_dict exception group 1: " + str(e))

    def test_hashNode(self):
        try:
            h = Hash("a.b.c", 42)
            node = h.getNode("a.b.c")
            self.assertEqual(node.getValue(), 42)

            self.assertEqual(node.getValueAs(Types.STRING), '42')
            self.assertEqual(node.getValueAs(Types.INT32), 42)

            self.assertEqual(node.getValueAs("STRING"), '42')
            self.assertEqual(node.getValueAs("INT32"), 42)
        except Exception as e:
            self.fail("test_node exception group 1: " + str(e))

        try:
            h = Hash("a.b.c", 42)
            h.setAttribute("a.b.c", "attr1", 15)
            node = h.getNode("a.b.c")
            self.assertEqual(node.getAttribute("attr1"), 15)

            self.assertEqual(node.getAttributeAs("attr1", Types.STRING), '15')
            self.assertEqual(node.getAttributeAs("attr1", Types.INT32), 15)

            self.assertEqual(node.getAttributeAs("attr1", "STRING"), '15')
            self.assertEqual(node.getAttributeAs("attr1", "INT32"), 15)
        except Exception as e:
            self.fail("test_node exception group 2: " + str(e))

        try:
            h = Hash("a.b.c", 15)
            node = h.getNode("a.b.c")
            self.assertEqual(node.getType(), Types.INT32)
            self.assertEqual(node.getValue(), 15)

            node.setType(Types.STRING)
            self.assertEqual(node.getType(), Types.STRING)
            self.assertEqual(node.getValue(), '15')

            node.setType("UINT32")
            self.assertEqual(node.getType(), Types.UINT32)
            self.assertEqual(node.getValue(), 15)
        except Exception as e:
            self.fail("test_node exception group 3: " + str(e))

    def test_hashAttributesNode(self):
        try:
            h = Hash("a.b.c", "value")
            h.setAttribute("a.b.c", "attr1", 10)
            h.setAttribute("a.b.c", "attr2", "test")
            attrs = h.getAttributes("a.b.c")

            node = attrs.getNode("attr1")
            self.assertEqual(node.getType(), Types.INT32)
            self.assertEqual(node.getValue(), 10)
            self.assertEqual(node.getValueAs(Types.STRING), '10')
            self.assertEqual(node.getValueAs("STRING"), '10')

            node.setType(Types.STRING)
            self.assertEqual(node.getType(), Types.STRING)
            self.assertEqual(node.getValue(), '10')
            self.assertEqual(node.getValueAs(Types.UINT32), 10)
            self.assertEqual(node.getValueAs("UINT32"), 10)

            node.setType("UINT32")
            self.assertEqual(node.getType(), Types.UINT32)
            self.assertEqual(node.getValue(), 10)

            node = attrs.getNode("attr2")
            self.assertEqual(node.getType(), Types.STRING)

        except Exception as e:
            self.fail("test_hashAttributesNode exception: " + str(e))

    def test_hash_update(self):
        h = Hash("a.b.a.b", 42)
        h2 = h.update({"a.b.a.b": 10})
        self.assertEqual(h["a.b.a.b"], 10)
        # test the correct reference
        self.assertTrue(h2 is h)
        del h
        gc.collect()
        self.assertEqual(h2["a.b.a.b"], 10)

        h = Hash()
        int_list = [1, 2, 3, 4]
        h.update(e=int_list)
        self.assertEqual(h["e"], int_list)
        self.assertEqual(h["e"][0], 1)
        int_list[0] = 10
        self.assertEqual(h["e"][0], 1)
        del int_list
        self.assertEqual(h["e"], [1, 2, 3, 4])
        with self.assertRaises(RuntimeError):
            h.update([1, 2, 3])

    def test_hashAttributes(self):
        try:
            h = Hash("a.b.c", "value")
            h.setAttribute("a.b.c", "attr1", 12)
            h.setAttribute("a.b.c", "attr2", True)

            attrs = h.getAttributes("a.b.c")

            self.assertEqual(attrs.getAs("attr1", Types.UINT32), 12)
            self.assertEqual(attrs.getAs("attr1", Types.STRING), '12')
            self.assertEqual(attrs.getAs("attr2", Types.BOOL), True)

            self.assertEqual(attrs.getAs("attr1", "UINT32"), 12)
            self.assertEqual(attrs.getAs("attr1", "STRING"), '12')
            self.assertEqual(attrs.getAs("attr2", "BOOL"), True)

        except Exception as e:
            self.fail("test_hashAttributes exception: " + str(e))


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import Hash, Types


def test_getSet_1():
    h = Hash()
    h.set("a.b.c1.d", 1)

    assert h.get("a").has("b")
    assert h.get("a.b").has("c1")
    assert h.get("a.b.c1").has("d")
    assert h.get("a.b.c1.d") == 1
    assert h.has("a.b.c1.d")
    assert "a.b.c1.d" in h
    assert h.get("a").has("b.c1")
    assert "b.c1" in h["a"]

    h.set("a.b.c2.d", 2.0)

    assert h.get("a.b").has("c2.d")
    assert h.get("a.b").has("c1.d")
    assert h.get("a.b.c1.d") == 1
    assert h.get("a.b.c2.d") == 2.0

    h.set("a.b[0]", Hash("a", 1))

    assert h.get("a").has("b")
    assert h["a"].has("b")
    assert "b" in h["a"]
    assert len(h.get("a")) == 1
    assert len(h["a"]) == 1
    assert len(h.get("a.b")) == 1
    assert len(h["a.b"]) == 1
    assert len(h.get("a.b")[0]) == 1
    assert len(h["a.b"][0]) == 1
    assert h["a.b"][0]["a"] == 1
    assert h["a.b[0].a"] == 1
    assert h.get("a.b")[0].get("a") == 1
    assert h.get("a.b[0].a") == 1

    h.set("a.b[2]", Hash("a", 1))

    assert h.get("a").has("b")
    assert len(h["a"]) == 1
    assert len(h["a.b"]) == 3
    assert h["a.b[0].a"] == 1
    assert h["a.b[2].a"] == 1
    assert h["a.b"][0]["a"] == 1
    assert h["a.b"][2]["a"] == 1
    assert h["a.b"][1].empty()


def test_getSet_2():
    h = Hash()
    h["a.b.c"] = 1
    h["a.b.c"] = 2

    assert h["a.b.c"] == 2
    assert h.has("a.b")
    assert not h.has("a.b.c.d")


def test_getSet_3():
    h = Hash("a[0]", Hash("a", 1), "a[1]", Hash("a", 1))

    assert h["a[0].a"] == 1
    assert h["a[1].a"] == 1
    assert h["a"][0]["a"] == 1
    assert h["a"][1]["a"] == 1


def test_getSet_4():
    h = Hash()
    h["x[0].y[0]"] = Hash("a", 4.2, "b", "red", "c", True)
    h["x[1].y[0]"] = Hash("a", 4.0, "b", "green", "c", False)

    assert h["x[0].y[0].c"]
    assert not h["x[1].y[0].c"]
    assert h["x[0].y[0].b"] == "red"
    assert h["x[1].y[0].b"] == "green"


def test_getSet_5():
    h1 = Hash("a[0].b[0]", Hash("a", 1))
    h2 = Hash("a[0].b[0]", Hash("a", 2))

    assert h1["a[0].b[0].a"] == 1
    h1["a[0]"] = h2
    assert h1["a[0].a[0].b[0].a"] == 2
    h1["a"] = h2
    assert h1["a.a[0].b[0].a"] == 2


def test_getSet_6():
    h = Hash()
    b = True
    h["a"] = b

    assert str(h.getType("a")) == "BOOL"
    assert h.getType("a") == Types.BOOL



import json

from karabo.bound import State
from karabo.bound.testing import ServerContext, eventLoop, sleepUntil


def test_server_context(eventLoop: eventLoop):
    config = {"PropTestA": {"classId": "PropertyTest"}}
    initA = json.dumps(config)
    serverA = ServerContext("testServerA",
                            ["log.level=DEBUG", f"init={initA}"])
    with serverA:
        remote = serverA.remote()
        sleepUntil(lambda: "PropTestA" in remote.getDevices(), timeout=10)
        assert len(remote.getDevices()) > 0
        assert remote.get("PropTestA", "state") == State.NORMAL
        remote.execute("PropTestA", "startWritingOutput")
        sleepUntil(lambda: remote.get("PropTestA", "state") == State.STARTED,
                   timeout=5)
        assert remote.get("PropTestA", "state") == State.STARTED

        config = {"PropTestB": {"classId": "PropertyTest"}}
        initB = json.dumps(config)

        serverB = ServerContext("testServerB",
                                ["log.level=DEBUG", f"init={initB}"],
                                client=remote)
        with serverB:
            sleepUntil(lambda: "PropTestB" in remote.getDevices(), timeout=10)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import copy

import numpy as np
import pytest

from karabo.bound import (
    ALARM_ELEMENT, BOOL_ELEMENT, BYTEARRAY_ELEMENT, DOUBLE_ELEMENT,
    FLOAT_ELEMENT, IMAGEDATA_ELEMENT, INT32_ELEMENT, INT64_ELEMENT,
    KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS, NDARRAY_ELEMENT,
    NODE_ELEMENT, OVERWRITE_ELEMENT, SLOT_ELEMENT, STATE_ELEMENT,
    STRING_ELEMENT, TABLE_ELEMENT, UINT32_ELEMENT, UINT64_ELEMENT,
    VECTOR_BOOL_ELEMENT, VECTOR_DOUBLE_ELEMENT, VECTOR_INT32_ELEMENT,
    VECTOR_STRING_ELEMENT, VECTOR_UINT32_ELEMENT, AccessLevel, AccessType,
    ArchivePolicy, AssemblyRules, Assignment, Encoding, Hash, MetricPrefix,
    NodeType, Schema, Types, Unit, cppNDArray, cppNDArrayCopy, fullyEqual)
from karabo.common.alarm_conditions import AlarmCondition
from karabo.common.states import State


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("ShapeX", "1.0")
class ShapeX:
    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected)
            .key("shadowEnabled")
            .description("Shadow enabled")
            .displayedName("Shadow")
            .assignmentOptional()
            .defaultValue(False)
            .init()
            .commit(),
        )


@KARABO_CLASSINFO("CircleX", "1.0")
class CircleX(ShapeX):
    @staticmethod
    def expectedParameters(expected):
        (
            FLOAT_ELEMENT(expected)
            .key("radius")
            .alias(1)
            .description("The radius of the circle")
            .displayedName("Radius")
            .minExc(0)
            .maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(50)
            .init()
            .commit(),

            STATE_ELEMENT(expected).key("state")
            .options(State.OFF, State.ON, State.ERROR)
            .commit(),

            STATE_ELEMENT(expected).key("stateN")
            .initialValue(State.NORMAL)
            .commit(),

            STATE_ELEMENT(expected).key("stateE")
            .initialValue(State.ERROR)
            .commit(),

            ALARM_ELEMENT(expected).key("alarm")
            .commit(),

            ALARM_ELEMENT(expected).key("alarmW")
            .initialValue(AlarmCondition.WARN)
            .commit(),

            ALARM_ELEMENT(expected).key("alarmA")
            .initialValue(AlarmCondition.ALARM)
            .commit(),

            STRING_ELEMENT(expected).key("status")
            .readOnly()
            .commit(),
        )


@KARABO_CLASSINFO("EditableCircleX", "1.0")
class EditableCircleX(CircleX):
    @staticmethod
    def expectedParameters(expected):
        CircleX.expectedParameters(expected)
        (
            OVERWRITE_ELEMENT(expected).key("radius")
            .setNewDisplayedName("New Radius")
            .setNewAlias(21)
            .setNewTags("operation,control")
            .setNewAssignmentMandatory()
            .setNowReconfigurable()
            .setNewUnit(Unit.METER)
            .setNewMetricPrefix(MetricPrefix.KILO)
            .setNewMinInc(12)
            .setNewMaxExc(120)
            .setNewAllowedStates(State.NORMAL)
            .setNowExpertAccess()
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("state")
            .setNewOptions(State.INIT, State.ERROR, State.NORMAL)
            .setNewDefaultValue(State.INIT)
            .commit(),

            OVERWRITE_ELEMENT(expected).key("status")
            .setNewOptions("a,b,c")
            .setNewDefaultValue("a")
            .commit(),

            OVERWRITE_ELEMENT(expected).key("alarm")
            .setNewDescription("The new alarm description")
            .setNewAssignmentOptional()
            .setNewDefaultValue(AlarmCondition.WARN)
            .setNowOperatorAccess()
            .commit(),

            OVERWRITE_ELEMENT(expected).key("alarmA")
            .setNowObserverAccess()
            .commit(),

        )


@KARABO_CLASSINFO("RectangleX", "1.0")
class RectangleX(ShapeX):
    @staticmethod
    def expectedParameters(expected):
        (
            FLOAT_ELEMENT(expected)
            .key("a")
            .alias(1)
            .description("Length of a")
            .displayedName("A")
            .minExc(0)
            .maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(50)
            .expertAccess()
            .init()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("b")
            .alias(1)
            .description("Length of b")
            .displayedName("B")
            .minExc(0)
            .maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(50)
            .init()
            .commit(),
        )


@KARABO_CLASSINFO("EditableRectangleX", "1.0")
class EditableRectangleX(RectangleX):
    @staticmethod
    def expectedParameters(expected):
        RectangleX.expectedParameters(expected)
        (
            OVERWRITE_ELEMENT(expected).key("a")
            .setNewAssignmentInternal()
            .setNewMinExc(30)
            .setNewMaxInc(60)
            .setNowExpertAccess()
            .commit(),

            OVERWRITE_ELEMENT(expected).key("b")
            .setNowReadOnly().setNewDefaultValue(12)
            .setNowOperatorAccess()
            .commit(),

        )


class GraphicsRendererX1:
    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected)
            .key("antiAlias")
            .tags("prop")
            .displayedName("Use Anti-Aliasing")
            .description("You may switch of for speed")
            .assignmentOptional()
            .defaultValue(True)
            .init()
            .expertAccess()
            .commit(),

            STRING_ELEMENT(expected)
            .key("color")
            .tags("prop")
            .displayedName("Color")
            .description("The default color for any shape")
            .assignmentOptional()
            .defaultValue("red")
            .reconfigurable()
            .allowedStates(State.OFF)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("bold")
            .tags("prop")
            .displayedName("Bold")
            .description("Toggles bold painting")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("shapes")
            .assignmentOptional()
            .defaultValue("circle")
            .commit(),

            NODE_ELEMENT(expected)
            .key("circle")
            .tags("shape")
            .displayedName("Circle")
            .description("A circle")
            .appendParametersOf(CircleX)
            .commit(),

            NODE_ELEMENT(expected)
            .key("rectangle")
            .tags("shape")
            .displayedName("Rectangle")
            .description("A rectangle")
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("rectangle.b")
            .description("Rectangle side - b")
            .displayedName("Side B")
            .tags("b")
            .assignmentOptional()
            .defaultValue(10)
            .init()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("rectangle.c")
            .description("Rectangle side - c")
            .displayedName("Side C")
            .assignmentOptional()
            .defaultValue(10)
            .init()
            .commit(),

            NODE_ELEMENT(expected)
            .key("triangle")
            .displayedName("triangle")
            .description("A triangle (Node element containing no other "
                         "elements)")
            .commit(),
        )


class TestStruct1:
    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected)
            .key("exampleKey1")
            .tags("hardware, poll")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional()
            .defaultValue("Navigation")
            .operatorAccess()
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected)
            .key("exampleKey2")
            .alias(10)
            .tags("hardware, poll")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5)
            .maxInc(25)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(10)
            .operatorAccess()
            .init()
            .commit(),

            UINT32_ELEMENT(expected)
            .key("exampleKey3")
            .alias(5.5)
            .tags("hardware, set")
            .displayedName("Example key 3")
            .description("Example key 3 description")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .minExc(10)
            .maxExc(20)
            .assignmentMandatory()
            .expertAccess()
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("exampleKey4")
            .alias("exampleAlias4")
            .tags("software")
            .displayedName("Example key 4")
            .description("Example key 4 description")
            .options("1.11     -2.22 5.55")
            .expertAccess()
            .minExc(-2.22)
            .maxExc(5.55)
            .assignmentInternal()
            .noDefaultValue()
            .commit(),
        )

        (
            # Always readOnly(), by default as well
            BYTEARRAY_ELEMENT(expected)
            .key("rarray")
            .alias("aliasReadArray")
            .tags("software")
            .displayedName("Example read array")
            .description("Example of ByteArray for reading")
            .operatorAccess()
            .readOnly()
            .initialValue(bytes('abcdef привет 012345', 'u8'))
            .commit(),

            BYTEARRAY_ELEMENT(expected)
            .key("rarray1")
            .displayedName("ReadArrayStr")
            .expertAccess()
            .readOnly()
            .initialValue('hello world')
            .commit(),

            BYTEARRAY_ELEMENT(expected)
            .key("rarray2")
            .displayedName("ReadByteArray")
            .observerAccess()
            .readOnly()
            .initialValue(bytearray('Tschüß!', 'u8'))
            .commit(),
        )

        (
            INT64_ELEMENT(expected)
            .key("exampleKey5")
            .alias("exampleAlias5")
            .tags("h/w; d.m.y", ";")
            .displayedName("Example key 5")
            .description("Example key 5 description")
            .setSpecialDisplayType("Int64DisplayType")
            .readOnly()
            .initialValue(1442244)
            .commit(),

            UINT64_ELEMENT(expected)
            .key("exampleKey5b")
            .readOnly()
            .defaultValue(42)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("exampleKey6")
            .alias([0x00123456, 0x0000aacc])
            .displayedName("Example key 6")
            .description("Example key 6 description")
            .readOnly()
            .initialValue(1.11)
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("exampleKey7")
            .displayedName("Example key 7")
            .allowedStates(State.STARTED, State.NORMAL)
            .readOnly()
            .initialValue([1, 2, 3])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("exampleKey7b")
            .readOnly()
            .initialValue([11, 22, 33])
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected)
            .key("exampleKey8")
            .readOnly()
            .initialValue([1.1, 2.2, 3.3])
            .archivePolicy(ArchivePolicy.NO_ARCHIVING)
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("exampleKey9")
            .setSpecialDisplayType("TestDisplayType")
            .readOnly()
            .initialValue(["Hallo", "World"])
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("vectInt")
            .readOnly()
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("exampleKey10")
            .displayedName("Example key 10")
            .minSize(2)
            .maxSize(7)
            .assignmentOptional()
            .defaultValueFromString("10, 20, 30")
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("exampleKey11")
            .displayedName("Example key 11")
            .assignmentOptional()
            .defaultValue([10, 20, 30])
            .observerAccess()
            .reconfigurable()
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected)
            .key("exampleKey12")
            .assignmentOptional()
            .defaultValueFromString("1.1, -2.2, 3.3")
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("exampleKey14")
            .assignmentOptional()
            .defaultValue(["Hallo", "World", "Test"])
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("exampleKey15")
            .assignmentOptional()
            .defaultValueFromString("word1, word2, test")
            .reconfigurable()
            .commit(),

            NDARRAY_ELEMENT(expected).key("exampleKey16")
            .dtype("UINT32")
            .shape([2, 5, 0])
            .commit(),

            INT32_ELEMENT(expected)
            .key("exampleIntKey")
            .assignmentOptional()
            .defaultValueFromString("20")
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(expected)
            .key("exampleBitsKey1")
            .tags("hardware")
            .displayedName("Example bits key 1")
            .description("Example bits key 1 description")
            .reconfigurable()
            .bin()
            .assignmentOptional()
            .defaultValue(0xdeadbeef)
            .commit(),

            UINT64_ELEMENT(expected)
            .key("exampleBitsKey2")
            .tags("hardware")
            .displayedName("Example bits key 2")
            .description("Example bits key 2 description")
            .reconfigurable()
            .bin("10:In Error, 21:Busy, 35:HV On, 55:Crate On")
            .assignmentOptional()
            .defaultValue(0xdeadbeefdeadface)
            .commit(),

            UINT32_ELEMENT(expected)
            .key("exampleBitsKey3")
            .tags("hardware")
            .displayedName("Example bits key 3")
            .description("Example bits key 3 description")
            .reconfigurable()
            .oct()
            .assignmentOptional()
            .defaultValue(0xbeefface)
            .commit(),

            UINT32_ELEMENT(expected)
            .key("exampleBitsKey4")
            .tags("hardware")
            .displayedName("Example bits key 4")
            .description("Example bits key 4 description")
            .reconfigurable()
            .hex()
            .assignmentOptional()
            .defaultValue(0xbeefface)
            .commit(),

            STRING_ELEMENT(expected)
            .key("testPath")
            .alias(5)
            .displayedName("Filename")
            .options("file1, file2, karabo.log")
            .assignmentOptional()
            .defaultValue("karabo.log")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testPath2")
            .readOnly()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testPath3")
            .assignmentMandatory()
            .commit(),

            SLOT_ELEMENT(expected)
            .key("slotTest")
            .alias(0x12345678)
            .tags("hard, soft")  # [...] & (...) supported only in karabind
            .displayedName("Reset")
            .description("Test slot element")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .operatorAccess()
            .commit(),

            SLOT_ELEMENT(expected)
            .key("slotClean")
            .alias("aliasClean")
            .displayedName("Clean")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .observerAccess()
            .commit(),

            SLOT_ELEMENT(expected)
            .key("slotStart")
            .alias("aliasStart")
            .tags('hard; soft')
            .displayedName("Start")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .expertAccess()
            .commit(),

            SLOT_ELEMENT(expected)
            .key("slotStop")
            .alias("aliasStop")
            .displayedName("Stop")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .expertAccess()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testChoice1")
            .displayedName("1. Select CircleX or RectangleX")
            .options("CircleX, RectangleX")
            .assignmentOptional().defaultValue("RectangleX")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testChoice2")
            .displayedName("2. Choose from menu")
            .options("CircleX, RectangleX")
            .assignmentOptional().noDefaultValue()
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testChoice3")
            .displayedName("3. Choose from menu")
            .options("CircleX, RectangleX")
            .assignmentMandatory()
            .init()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testChoice4")
            .displayedName("4. 3Choose from menu")
            .options("CircleX, RectangleX")
            .assignmentOptional().defaultValue("CircleX")
            .reconfigurable()
            .commit(),

            NODE_ELEMENT(expected)
            .key("CircleX")
            .appendParametersOfConfigurableClass(ShapeX, "CircleX")
            .commit(),

            NODE_ELEMENT(expected)
            .key("RectangleX")
            .appendParametersOfConfigurableClass(ShapeX, "RectangleX")
            .commit(),

        )

        row = Schema()
        (
            INT32_ELEMENT(row)
            .key("int")
            .readOnly()
            .initialValue(0)
            .commit(),

            TABLE_ELEMENT(expected)
            .key("tableI")
            .setColumns(row)
            .readOnly()
            .initialValue([Hash("int", 2)])
            .commit(),

            TABLE_ELEMENT(expected)
            .key("tableD")
            .setColumns(row)
            .readOnly()
            .defaultValue([Hash("int", 3)])
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("myNode")
            .description("WidgetNode Test")
            .setSpecialDisplayType("WidgetNode")
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("OtherSchemaElementsX", "1.0")
class OtherSchemaElementsX:
    @staticmethod
    def expectedParameters(expected):
        (
            SLOT_ELEMENT(expected)
            .key("slotTest")
            .displayedName("Reset")
            .description("Test slot element")
            .allowedStates(State.STARTED, State.STOPPED, State.ERROR)
            .commit(),

            STRING_ELEMENT(expected)
            .description("File name")
            .key("filename")
            .alias(5)
            .displayedName("Filename")
            .options("file1, file2, karabo.log")
            .assignmentOptional()
            .defaultValue("karabo.log")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("testfile")
            # .isInputFile()
            .readOnly()
            .defaultValue("initFile")  # == initialValue
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("vecIntReconfig")
            .assignmentOptional()
            .defaultValue([10, 20, 30])
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("vecIntReconfigStr")
            .assignmentOptional()
            .defaultValueFromString("11, 22, 33")
            .reconfigurable()
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected)
            .key("vecDoubleReconfigStr")
            .assignmentOptional()
            .defaultValueFromString("1.1, 2.2, 3.3")
            .reconfigurable()
            .commit(),

            VECTOR_BOOL_ELEMENT(expected)
            .key("vecBool")
            .tags("h/w; d.m.y", ";")
            .allowedStates(State.STARTED, State.STOPPED)
            .minSize(2)
            .maxSize(7)
            .assignmentMandatory()
            .commit(),

            NDARRAY_ELEMENT(expected)
            .key("arrBool")
            .dtype(Types.BOOL)
            .shape([3, 2])
            .commit(),

            NDARRAY_ELEMENT(expected)
            .key("arrInt8")
            .dtype(Types.INT8)
            .shape([3, 2])
            .commit(),

            NDARRAY_ELEMENT(expected)
            .key("arrUInt16")
            .dtype(Types.UINT16)
            .shape([3, 2])
            .commit(),

            NDARRAY_ELEMENT(expected)
            .key("arrFloat")
            .dtype(Types.FLOAT)
            .shape([3, 2])
            .commit(),

            # Keep shape as string to also test that
            NDARRAY_ELEMENT(expected)
            .key("arrDouble")
            .dtype(Types.DOUBLE)
            .shape("3,2,-1")
            .commit(),

            NDARRAY_ELEMENT(expected)
            .key("arrUndefined")
            # leaf type undefined
            .shape("0,3,0")
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("image")
            .commit(),
        )

        rowSchema = Schema()
        (
            INT32_ELEMENT(rowSchema)
            .key("a")
            .minInc(2)
            .maxExc(10)
            .displayedName("A")
            .assignmentOptional()
            .defaultValue(2)
            .commit(),

            STRING_ELEMENT(rowSchema)
            .key("b")
            .assignmentOptional()
            .noDefaultValue()
            .commit(),

            FLOAT_ELEMENT(rowSchema)
            .key("c")
            .assignmentOptional()
            .noDefaultValue()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("testTable")
            .alias([1, 2])
            .tags("hard,soft")
            .description("Test table description")
            .displayedName("Test table")
            .setColumns(rowSchema)
            .maxSize(10)
            .minSize(2)
            .assignmentOptional()
            .defaultValue([Hash("a", 3, "b", "foo", "c", 3.1415),
                           Hash("a", 3, "b", "foo", "c", 2.78)])
            .expertAccess()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("testTableEmptyDefault")
            .alias(3)
            .tags("soft")
            .setColumns(rowSchema)
            .assignmentOptional()
            .defaultValue([])
            .expertAccess()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("testTableNoDefault")
            .alias("aliasForTestTableNoDefault")
            .tags("hard")
            .setColumns(rowSchema)
            .assignmentOptional()
            .noDefaultValue()
            .observerAccess()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("tableMandatory")
            .alias("aliasForTableMandatory")
            .tags("hard")
            .setColumns(rowSchema)
            .assignmentMandatory()
            .operatorAccess()
            .init()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("tableInternal")
            .alias("aliasForTableInternal")
            .tags("hard")
            .description("Test table with assignmentInternal")
            .setColumns(rowSchema)
            .reconfigurable()
            .operatorAccess()
            .assignmentInternal()
            .noDefaultValue()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("tableReadOnly")
            .alias("aliasForTableReadOnly")
            .tags("exp,hard,soft")
            .setColumns(rowSchema)
            .allowedStates(State.STARTED, State.STOPPED)
            .operatorAccess()
            .readOnly()
            .initialValue([Hash("a", 22, "b", "foo22", "c", 1.234),
                           Hash("a", 33, "b", "foo33", "c", 2.345),
                           Hash("a", 44, "b", "foo44", "c", 3.456)])
            .archivePolicy(ArchivePolicy.EVERY_EVENT)
            .commit(),
        )


class SomeClass:
    @staticmethod
    def expectedParameters(expected):
        (
            INT32_ELEMENT(expected)
            .key("x")
            .alias(10)
            .tags("IK,BH")
            .displayedName("Xkey")
            .description("Example of X key description")
            .options("5, 25, 10")
            .minInc(5)
            .maxInc(25)
            .unit(Unit.AMPERE)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional()
            .defaultValue(5)
            .init()
            .expertAccess()
            .commit(),

            INT32_ELEMENT(expected)
            .key("y")
            .alias("bla")
            .tags("CY")
            .displayedName("Ykey")
            .description("Example of Y key description")
            .options("5, 25, 10")
            .minExc(0)
            .maxExc(29)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.CENTI)
            .assignmentOptional()
            .defaultValue(10)
            .init()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("a")
            .readOnly()
            .initialValue(1.11)
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("somelist")
            .displayedName("Ykey")
            .description("Example of Y key description")
            .reconfigurable()
            .assignmentOptional()
            .defaultValue([])
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("imageSetDims")
            .displayedName("Image setDimensions")
            .description("ImageData element: setDimensions")
            .setDimensions("110, 210")
            .operatorAccess()
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("imageSetType")
            .displayedName("Image setType")
            .description("ImageData Element: setType('UINT16')")
            .setDimensions("110, 210")
            .setType('UINT16')
            .observerAccess()
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("imageSetEncoding")
            .displayedName("Image setEncoding")
            .description("ImageData Element: setEncoding('RGBA')")
            .setDimensions("300, 200")
            .setType(Types.UINT32)
            .setEncoding("RGBA")
            .operatorAccess()
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("imageSetGeometry")
            .displayedName("ImageData setGeometry")
            .description("ImageData Element: setGeometry")
            .setDimensions("500, 500")
            .setEncoding(Encoding.GRAY)
            # .setGeometry()
            .expertAccess()
            .commit(),

            IMAGEDATA_ELEMENT(expected)
            .key("myImageElement")
            .setDimensions("500, 800")
            .expertAccess()
            .commit(),
        )


def test_buildUp():
    schema = ShapeX.getSchema("CircleX", AssemblyRules())
    states = [State.OFF.value, State.ON.value, State.ERROR.value]
    assert schema.getOptions("state") == states
    assert schema.isAccessInitOnly("shadowEnabled") is True
    assert schema.isAccessInitOnly("radius") is True
    assert schema.isLeaf("radius") is True
    assert schema.getDisplayedName("radius") == "Radius"
    assert schema.getDescription("radius") == "The radius of the circle"
    assert schema.getAliasFromKey("radius") == 1
    assert schema.getAliasAsString("radius") == "1"
    assert schema.hasTags("radius") is False
    assert schema.isAssignmentOptional("radius") is True
    assert schema.isAssignmentMandatory("radius") is False
    assert schema.isAccessReadOnly("radius") is False
    assert schema.isAccessReconfigurable("radius") is False
    assert schema.isAccessInitOnly("radius") is True
    assert schema.getUnitName("radius") == "meter"
    assert schema.getMetricPrefixName("radius") == "milli"
    assert schema.hasMinExc("radius") is True
    assert schema.hasMinInc("radius") is False
    assert schema.getMinExc("radius") == 0
    assert schema.getMaxExc("radius") == 100
    assert schema.getRequiredAccessLevel("radius") == AccessLevel.OPERATOR

    schema = Schema()
    EditableCircleX.expectedParameters(schema)
    allowedStates = schema.getOptions("state")
    assert allowedStates == ['INIT', 'ERROR', 'NORMAL']
    assert schema.getDefaultValue("state") == 'INIT'
    assert schema.getDefaultValue("stateN") == 'NORMAL'
    assert schema.getDefaultValue("stateE") == 'ERROR'
    assert schema.getDefaultValue("alarmW") == 'warn'
    assert schema.getDefaultValue("alarmA") == 'alarm'
    assert schema.getDefaultValue("alarm") == 'warn'
    assert schema.getDisplayedName("radius") == "New Radius"
    assert schema.getAliasFromKey("radius") == 21
    assert schema.getAliasAsString("radius") == "21"
    assert schema.getTags("radius") == ["operation,control"]
    assert schema.isAssignmentOptional("radius") is False
    assert schema.isAssignmentMandatory("radius") is True
    assert schema.isAccessReadOnly("radius") is False
    assert schema.isAccessReconfigurable("radius") is True
    assert schema.getUnitName("radius") == "meter"
    assert schema.getMetricPrefixName("radius") == "kilo"
    assert schema.hasMinExc("radius") is True
    assert schema.hasMinInc("radius") is True
    assert schema.getMinInc("radius") == 12
    assert schema.getMaxExc("radius") == 120
    assert schema.getRequiredAccessLevel("radius") == AccessLevel.EXPERT
    assert schema.getDescription("alarm") == "The new alarm description"
    assert schema.getRequiredAccessLevel("alarm") == AccessLevel.OPERATOR

    allowedStates = schema.getOptions("status")
    assert allowedStates == ['a', 'b', 'c']
    assert schema.getDefaultValue("status") == 'a'

    schema = Schema()
    EditableRectangleX.expectedParameters(schema)
    assert schema.getMinExc("a") == 30
    assert schema.getMaxInc("a") == 60
    assert schema.isAssignmentInternal("a") is True
    assert schema.isAssignmentOptional("a") is False
    assert schema.isAssignmentMandatory("a") is False
    assert schema.isAccessReadOnly("a") is False
    assert schema.isAccessReadOnly("b") is True
    assert schema.getRequiredAccessLevel("a") == AccessLevel.EXPERT
    assert schema.getRequiredAccessLevel("b") == AccessLevel.OPERATOR

    schema = Schema()
    GraphicsRendererX1.expectedParameters(schema)
    assert schema.isAccessInitOnly("circle.radius") is True
    assert schema.isLeaf("circle.radius") is True


def test_invalidNodes():
    schema = Schema()
    # Empty strings are forbidden as keys:
    with pytest.raises(RuntimeError):
        INT32_ELEMENT(schema).key("").description("Empty key is forbidden")

    # Also under a node:
    NODE_ELEMENT(schema).key("node").commit()
    with pytest.raises(RuntimeError):
        INT32_ELEMENT(schema).key("node.").description("Empty key forbidden")

    # Various characters are forbidden:
    forbidden = "()[]<>`~!#$%^&*-+=|\\:;'\",? "
    valid = "valid"
    for bad in forbidden:
        invalid = valid + bad
        with pytest.raises(RuntimeError):
            INT32_ELEMENT(schema).key(invalid)
        with pytest.raises(RuntimeError):
            INT32_ELEMENT(schema).key("node." + invalid)

    # Digits as first character are also forbidden
    for i in range(10):
        invalid = str(i) + valid  # test '0', '1', '2', ... one after another
        with pytest.raises(RuntimeError):
            INT32_ELEMENT(schema).key(invalid)
        with pytest.raises(RuntimeError):
            INT32_ELEMENT(schema).key("node." + invalid)

    # No '/' as first, neither
    invalid = "/invalid"
    with pytest.raises(RuntimeError):
        INT32_ELEMENT(schema).key(invalid)
    with pytest.raises(RuntimeError):
        INT32_ELEMENT(schema).key("node." + invalid)

    allValidCharacters = ("abcdefghijklmnopqrstuvwxyz/"  # tolerate '/'
                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                          "0123456789_")
    INT32_ELEMENT(schema).key(allValidCharacters).readOnly().commit()
    assert allValidCharacters in schema

    # Check tolerance if requested to be not strict
    # Test various _ELEMENT classes that have their own bindings
    # (NODE_ELEMENT is tested elsewhere)
    elements = [ALARM_ELEMENT, IMAGEDATA_ELEMENT, INT32_ELEMENT,
                NDARRAY_ELEMENT, SLOT_ELEMENT, STATE_ELEMENT, TABLE_ELEMENT,
                VECTOR_DOUBLE_ELEMENT]
    for ELEMENT in elements:
        tolerated = "valid@"
        ELEMENT(schema).key(tolerated, False)
        with pytest.raises(RuntimeError,
                           match="illegal character at position 5"):
            ELEMENT(schema).key(tolerated)

        # Now with two tolerated characters, one at the end
        tolerated = "valid-more@"
        ELEMENT(schema).key(tolerated, False)
        with pytest.raises(RuntimeError,
                           match="illegal character at position 5"):
            ELEMENT(schema).key(tolerated)

        invalid = "totally&invalid"
        with pytest.raises(RuntimeError,
                           match="not tolerated character at position 7"):
            ELEMENT(schema).key(invalid, False)
        with pytest.raises(RuntimeError,
                           match="illegal character at position 7"):
            ELEMENT(schema).key(invalid, True)


def test_getRootName():
    schema = Schema("MyTest")
    TestStruct1.expectedParameters(schema)
    assert schema.getRootName() == "MyTest"


def test_getTags():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getTags("exampleKey1")[0] == "hardware"
    assert schema.getTags("exampleKey1")[1] == "poll"
    assert schema.getTags("exampleKey2")[0] == "hardware"
    assert schema.getTags("exampleKey2")[1] == "poll"
    assert schema.getTags("exampleKey3")[0] == "hardware"
    assert schema.getTags("exampleKey3")[1] == "set"
    assert schema.getTags("exampleKey4")[0] == "software"
    assert schema.getTags("exampleKey5")[0] == "h/w"
    assert schema.getTags("exampleKey5")[1] == "d.m.y"
    assert schema.getTags("rarray")[0] == "software"


def test_setTags():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.hasTags('x') is True
    assert schema.getTags('x') == ['IK', 'BH']
    schema.setTags('x', 'CY,SE')
    assert schema.getTags('x') == ['CY', 'SE']


def test_getsetAccessLevel():
    sch = Schema()
    SomeClass.expectedParameters(sch)
    assert sch.getRequiredAccessLevel('x') == AccessLevel.EXPERT
    assert sch.getRequiredAccessLevel('y') == AccessLevel.OPERATOR
    assert sch.getRequiredAccessLevel('a') == AccessLevel.OBSERVER

    sch.setRequiredAccessLevel('x', AccessLevel.EXPERT)
    sch.setRequiredAccessLevel('y', AccessLevel.OPERATOR)
    assert sch.getRequiredAccessLevel('x') == AccessLevel.EXPERT
    assert sch.getRequiredAccessLevel('y') == AccessLevel.OPERATOR

    sch = Schema()
    TestStruct1.expectedParameters(sch)
    assert sch.getRequiredAccessLevel('exampleKey1') == AccessLevel.OPERATOR
    assert sch.getRequiredAccessLevel('exampleKey2') == AccessLevel.OPERATOR
    assert sch.getRequiredAccessLevel('exampleKey3') == AccessLevel.EXPERT
    assert sch.getRequiredAccessLevel('exampleKey4') == AccessLevel.EXPERT
    # default for readOnly
    assert sch.getRequiredAccessLevel('exampleKey5') == AccessLevel.OBSERVER
    # default for reconfigurable
    assert sch.getRequiredAccessLevel('exampleKey10') == AccessLevel.OPERATOR
    # observerAccess in reconfigurable
    assert sch.getRequiredAccessLevel('exampleKey11') == AccessLevel.OBSERVER
    assert sch.getRequiredAccessLevel('rarray') == AccessLevel.OPERATOR


def test_isNode():
    schema = Schema("test")
    GraphicsRendererX1.expectedParameters(schema)
    assert schema.getRootName() == "test"
    assert schema.isNode("circle") is True
    assert schema.isNode("rectangle") is True
    assert schema.isNode("shapes") is False


def test_getNodeType():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    nodeType = schema.getNodeType("exampleKey1")
    assert nodeType == NodeType.LEAF
    assert schema.getNodeType("exampleKey5") == NodeType.LEAF
    schema = Schema()
    GraphicsRendererX1.expectedParameters(schema)
    assert schema.getNodeType("shapes") == NodeType.LEAF
    assert schema.getNodeType("circle") == NodeType.NODE
    assert schema.getNodeType("rectangle") == NodeType.NODE


def test_getValueType():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getValueType("exampleKey1") == Types.STRING
    assert schema.getValueType("exampleKey2") == Types.INT32
    assert schema.getValueType("exampleKey3") == Types.UINT32
    assert schema.getValueType("exampleKey4") == Types.DOUBLE
    assert schema.getValueType("exampleKey5") == Types.INT64
    assert schema.getValueType("exampleKey5b") == Types.UINT64
    assert schema.getValueType("exampleKey7") == Types.VECTOR_INT32
    assert schema.getValueType("exampleKey7b") == Types.VECTOR_UINT32
    assert schema.getValueType("exampleKey8") == Types.VECTOR_DOUBLE
    assert schema.getValueType("exampleKey9") == Types.VECTOR_STRING
    assert schema.getValueType("rarray") == Types.BYTE_ARRAY


def test_getAliasAsString():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getAliasAsString("exampleKey2") == "10"
    assert schema.getAliasAsString("exampleKey3") == "5.5"
    assert schema.getAliasAsString("exampleKey4") == "exampleAlias4"
    assert schema.getAliasAsString("exampleKey5") == "exampleAlias5"
    assert schema.getAliasAsString("exampleKey6") == "1193046,43724"
    assert schema.getAliasAsString("testPath") == "5"
    assert schema.getAliasAsString("rarray") == "aliasReadArray"


def test_keyHasAlias():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.keyHasAlias("exampleKey1") is False
    assert schema.keyHasAlias("exampleKey2") is True
    assert schema.keyHasAlias("exampleKey3") is True
    assert schema.keyHasAlias("exampleKey4") is True
    assert schema.keyHasAlias("exampleKey5") is True
    assert schema.keyHasAlias("exampleKey6") is True
    assert schema.keyHasAlias("testPath") is True
    assert schema.keyHasAlias("rarray") is True


def test_aliasHasKey():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.aliasHasKey(10) is True
    assert schema.aliasHasKey(5.5) is True
    assert schema.aliasHasKey("exampleAlias4") is True
    assert schema.aliasHasKey("exampleAlias5") is True
    assert schema.aliasHasKey([0x00123456, 0x0000aacc]) is True
    assert schema.aliasHasKey(7) is False
    assert schema.aliasHasKey(5) is True
    assert schema.aliasHasKey("aliasReadArray") is True


def test_getAliasFromKey():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getAliasFromKey("exampleKey2") == 10
    assert schema.getAliasFromKey("exampleKey3") == 5.5
    assert schema.getAliasFromKey("exampleKey4") == "exampleAlias4"
    assert schema.getAliasFromKey("exampleKey5") == "exampleAlias5"
    assert schema.getAliasFromKey("exampleKey6") == [0x00123456, 0x0000aacc]
    assert schema.getAliasFromKey("testPath") == 5
    assert schema.getAliasFromKey("rarray") == "aliasReadArray"


def test_setAlias():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.getAliasFromKey("x") == 10
    schema.setAlias('x', 'abc')
    assert schema.getAliasFromKey("x") == 'abc'
    schema.setAlias('x', 99)
    assert schema.getAliasFromKey("x") == 99


def test_getKeyFromAlias():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getKeyFromAlias(10) == "exampleKey2"
    assert schema.getKeyFromAlias(5.5) == "exampleKey3"
    assert schema.getKeyFromAlias("exampleAlias4") == "exampleKey4"
    assert schema.getKeyFromAlias("exampleAlias5") == "exampleKey5"
    assert schema.getKeyFromAlias([0x00123456, 0x0000aacc]) == "exampleKey6"
    assert schema.getKeyFromAlias(5) == "testPath"
    assert schema.getKeyFromAlias("aliasReadArray") == "rarray"


def test_getAccessMode():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getAccessMode("exampleKey1") == AccessType.WRITE
    assert schema.getAccessMode("exampleKey2") == AccessType.INIT
    assert schema.getAccessMode("exampleKey3") == AccessType.WRITE
    assert schema.getAccessMode("exampleKey4") == AccessType.INIT
    assert schema.getAccessMode("exampleKey5") == AccessType.READ
    assert schema.getAccessMode("exampleKey8") == AccessType.READ
    assert schema.getAccessMode("exampleKey10") == AccessType.WRITE
    assert schema.getAccessMode("testPath") == AccessType.WRITE
    assert schema.getAccessMode("testPath2") == AccessType.READ
    assert schema.getAccessMode("testPath3") == AccessType.INIT
    assert schema.getAccessMode("rarray") == AccessType.READ


def test_getAssignment():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getAssignment("exampleKey1") == Assignment.OPTIONAL
    assert schema.getAssignment("exampleKey2") == Assignment.OPTIONAL
    assert schema.getAssignment("exampleKey3") == Assignment.MANDATORY
    assert schema.getAssignment("exampleKey4") == Assignment.INTERNAL
    assert schema.getAssignment("exampleKey5") == Assignment.OPTIONAL
    assert schema.getAssignment("exampleKey8") == Assignment.OPTIONAL
    assert schema.getAssignment("exampleKey10") == Assignment.OPTIONAL
    assert schema.getAssignment("testPath") == Assignment.OPTIONAL
    assert schema.getAssignment("testPath2") == Assignment.OPTIONAL
    assert schema.getAssignment("testPath3") == Assignment.MANDATORY


def test_setAssignment():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.hasAssignment('x') is True
    assert schema.isAssignmentOptional('x') is True
    assert schema.isAssignmentMandatory('x') is False
    assert schema.getAssignment('x') == Assignment.OPTIONAL
    schema.setAssignment('x', Assignment.MANDATORY)
    assert schema.isAssignmentOptional('x') is False
    assert schema.isAssignmentMandatory('x') is True
    assert schema.getAssignment('x') == Assignment.MANDATORY


def test_getDescription():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getDescription('exampleKey1') == "Example key 1 description"
    assert schema.getDescription('rarray') == \
        "Example of ByteArray for reading"


def test_setDescription():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    schema.setDescription('exampleKey1', "No description")
    assert schema.getDescription('exampleKey1') == "No description"


def test_getOptions():
    schema = Schema()
    TestStruct1.expectedParameters(schema)

    options = schema.getOptions("exampleKey1")
    assert options[0] == "Radio"
    assert options[1] == "Air Condition"
    assert options[2] == "Navigation"

    assert schema.getOptions("exampleKey2")[0] == 5
    assert schema.getOptions("exampleKey2")[1] == 25
    assert schema.getOptions("exampleKey2")[2] == 10
    assert schema.getOptions("exampleKey4")[0] == 1.11
    assert schema.getOptions("exampleKey4")[1] == -2.22
    assert schema.getOptions("exampleKey4")[2] == 5.55

    assert schema.getOptions("testPath")[0] == "file1"
    assert schema.getOptions("testPath")[1] == "file2"


def test_setOptions():
    schema = Schema()
    SomeClass.expectedParameters(schema)

    options = schema.getOptions("x")
    assert options[0] == 5
    assert options[1] == 25
    assert options[2] == 10
    assert schema.getOptions("x") == [5, 25, 10]

    schema.setOptions('x', '20, 5, 11, 13, 25')
    options = schema.getOptions("x")
    assert options == [20, 5, 11, 13, 25]


def test_getDefaultValue():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getDefaultValue("exampleKey1") == "Navigation"
    assert schema.getDefaultValue("exampleKey2") == 10
    assert schema.getDefaultValueAs("exampleKey2", Types.STRING) == "10"
    assert schema.getDefaultValue("exampleKey5") == 1442244
    assert schema.getDefaultValueAs("exampleKey5", Types.STRING) == "1442244"
    assert schema.getDefaultValue("exampleKey6") == 1.11
    assert schema.getDefaultValueAs("exampleKey6", Types.DOUBLE) == 1.11
    assert schema.getDefaultValue("exampleKey7") == [1, 2, 3]
    assert schema.getDefaultValue("exampleKey8") == [1.1, 2.2, 3.3]
    assert schema.getDefaultValue("exampleKey9") == ["Hallo", "World"]

    # 'readOnly'-element that does not specify
    #  'initialValue' has 'defaultValue' equal to string "" :
    assert schema.getDefaultValue("testPath2") == ""
    # ... and for vectors equal to empty list:
    assert schema.getDefaultValue("vectInt") == []

    assert schema.getDefaultValue("exampleIntKey") == 20

    assert schema.getDefaultValue("exampleKey5") == 1442244

    assert schema.getDefaultValue("exampleKey7") == [1, 2, 3]

    # readOnly default specified by 'defaultValue'. not 'initialValue':
    assert schema.getDefaultValue("exampleKey5b") == 42
    assert schema.getDefaultValue("exampleKey7b") == [11, 22, 33]
    assert schema.getDefaultValue("rarray") == \
        b'abcdef \xd0\xbf\xd1\x80\xd0\xb8\xd0\xb2\xd0\xb5\xd1\x82 012345'
    assert schema.getDefaultValue("rarray1") == b'hello world'
    assert schema.getDefaultValue("rarray2") == b'Tsch\xc3\xbc\xc3\x9f!'


def test_setDefaultValue():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.isAssignmentOptional('x') is True
    assert schema.hasDefaultValue('x') is True
    assert schema.getDefaultValue("x") == 5
    schema.setDefaultValue("x", 10)
    assert schema.getDefaultValue("x") == 10


def test_getAllowedStates():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    astates = schema.getAllowedStates("exampleKey3")
    assert astates[0] == State.STARTED
    assert astates[1] == State.STOPPED
    assert astates[2] == State.NORMAL
    assert schema.getAllowedStates("exampleKey3")[2] == State.NORMAL
    assert schema.getAllowedStates("exampleKey7")[0] == State.STARTED
    assert schema.getAllowedStates("exampleKey7")[1] == State.NORMAL


def test_getUnit():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getUnit("exampleKey2") == Unit.METER
    assert schema.getUnitName("exampleKey2") == "meter"
    assert schema.getUnitSymbol("exampleKey2") == "m"


def test_setUnit():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.getUnit("x") == Unit.AMPERE
    schema.setUnit('x', Unit.METER)
    assert schema.getUnit("x") == Unit.METER
    assert schema.getUnit("x") == Unit.METER
    assert schema.getUnitName("x") == "meter"
    assert schema.getUnitSymbol("x") == "m"


def test_getMetricPrefix():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getMetricPrefix("exampleKey2") == MetricPrefix.MILLI
    assert schema.getMetricPrefixName("exampleKey2") == "milli"
    assert schema.getMetricPrefixSymbol("exampleKey2") == "m"


def test_setMetricPrefix():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.getMetricPrefix("x") == MetricPrefix.MILLI
    schema.setMetricPrefix("x", MetricPrefix.MICRO)
    assert schema.getMetricPrefix("x") == MetricPrefix.MICRO
    assert schema.getMetricPrefixName("x") == "micro"
    assert schema.getMetricPrefixSymbol("x") == "u"


def test_getMinIncMaxInc():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getMinInc("exampleKey2") == 5
    assert schema.getMinIncAs("exampleKey2", Types.STRING) == "5"
    assert schema.getMaxInc("exampleKey2") == 25
    assert schema.getMaxIncAs("exampleKey2", Types.STRING) == "25"


def test_setMinIncMaxInc():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.getMinInc("x") == 5
    assert schema.getMinIncAs("x", Types.STRING) == "5"
    assert schema.getMaxInc("x") == 25
    assert schema.getMaxIncAs("x", Types.STRING) == "25"
    schema.setMinInc('x', 3)
    schema.setMaxInc('x', 30)
    assert schema.getMinInc("x") == 3
    assert schema.getMinIncAs("x", Types.STRING) == "3"
    assert schema.getMaxInc("x") == 30
    assert schema.getMaxIncAs("x", Types.STRING) == "30"


def test_getMinExcMaxExc():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.getMinExc("exampleKey3") == 10
    assert schema.getMinExc("exampleKey4") == -2.22
    assert schema.getMaxExc("exampleKey3") == 20
    assert schema.getMaxExc("exampleKey4") == 5.55
    assert schema.getMinExcAs("exampleKey3", Types.STRING) == "10"
    assert schema.getMinExcAs("exampleKey4", Types.STRING) == "-2.22"
    assert schema.getMaxExcAs("exampleKey3", Types.STRING) == "20"
    assert schema.getMaxExcAs("exampleKey4", Types.STRING) == "5.55"


def test_setMinExcMaxExc():
    schema = Schema()
    SomeClass.expectedParameters(schema)
    assert schema.getMinExc("y") == 0
    assert schema.getMaxExc("y") == 29
    schema.setMinExc("y", 2)
    schema.setMaxExc("y", 30)
    assert schema.getMinExc("y") == 2
    assert schema.getMaxExc("y") == 30


def test_hasgetsetMinMaxSize():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    assert schema.hasMinSize('exampleKey10') is True
    assert schema.hasMaxSize('exampleKey10') is True
    assert schema.getMinSize('exampleKey10') == 2
    assert schema.getMaxSize('exampleKey10') == 7
    schema.setMinSize('exampleKey10', 5)
    schema.setMaxSize('exampleKey10', 12)
    assert schema.getMinSize('exampleKey10') == 5
    assert schema.getMaxSize('exampleKey10') == 12
    # ------
    assert schema.hasMinSize('exampleKey11') is False
    assert schema.hasMaxSize('exampleKey11') is False
    schema.setMinSize('exampleKey11', 1)
    schema.setMaxSize('exampleKey11', 42)
    assert schema.hasMinSize('exampleKey11') is True
    assert schema.hasMaxSize('exampleKey11') is True
    assert schema.getMinSize('exampleKey11') == 1
    assert schema.getMaxSize('exampleKey11') == 42


def test_vectorElement():
    s = Schema()
    TestStruct1.expectedParameters(s)
    assert s.isAccessReadOnly("exampleKey7") is True
    assert s.hasDefaultValue("exampleKey7") is True
    assert s.hasDefaultValue("exampleKey10") is True
    assert s.hasDefaultValue("exampleKey11") is True
    assert s.hasDefaultValue("exampleKey12") is True
    assert s.hasDefaultValue("exampleKey14") is True
    assert s.hasDefaultValue("exampleKey15") is True
    assert s.getDefaultValue("exampleKey10") == [10, 20, 30]
    assert s.getDefaultValue("exampleKey12") == [1.1, -2.2, 3.3]
    assert s.getDefaultValue("exampleKey11") == [10, 20, 30]
    assert s.getDefaultValueAs("exampleKey11", Types.STRING) == "10,20,30"
    assert s.getDefaultValueAs("exampleKey11",
                               Types.VECTOR_INT32) == [10, 20, 30]
    assert s.getDefaultValue("exampleKey14") == ["Hallo", "World", "Test"]
    assert s.getDefaultValueAs("exampleKey14",
                               Types.STRING) == "Hallo,World,Test"
    assert s.getDefaultValue("exampleKey15") == ["word1", "word2", "test"]
    assert s.getDefaultValueAs("exampleKey15",
                               Types.STRING) == "word1,word2,test"
    assert s.getMinSize("exampleKey10") == 2
    assert s.getMaxSize("exampleKey10") == 7


def test_table_element():
    schema = Schema()
    TestStruct1.expectedParameters(schema)
    # Both, defaultValue and initialValue should work for read-only tables
    default = schema.getDefaultValue("tableI")
    assert len(default) == 1
    assert fullyEqual(default[0], Hash("int", 2)) is True
    default = schema.getDefaultValue("tableD")
    assert len(default) == 1
    assert fullyEqual(default[0], Hash("int", 3)) is True

    # There are more tests of tables in the integration tests...


def test_overwrite_restrictions_for_options():
    schema = Schema()
    (
        INT32_ELEMENT(schema).key("range")
        .displayedName("Range")
        .options("0,1")
        .assignmentOptional().defaultValue(0)
        .reconfigurable()
        .commit()
    )
    vec = schema.getOptions("range")
    assert len(vec) == 2
    assert vec[0] == 0
    assert vec[1] == 1
    (
        OVERWRITE_ELEMENT(schema).key("range")
        .setNewOptions("0,1,2")
        .commit()
    )
    vec = schema.getOptions("range")
    assert len(vec) == 3
    assert vec[0] == 0
    assert vec[1] == 1
    assert vec[2] == 2


def test_overwrite_tags():
    schema = Schema()
    (
        INT32_ELEMENT(schema).key("taggedProp")
        .tags("bip, bop")
        .assignmentOptional().defaultValue(0)
        .reconfigurable()
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "bip"
    assert vec[1] == "bop"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags("doff")
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 1
    assert vec[0] == "doff"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags(["doff", "deff"])
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "doff"
    assert vec[1] == "deff"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags(("chip", "chop"))
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "chip"
    assert vec[1] == "chop"
    # Only iterable of str are allowed
    with pytest.raises(RuntimeError):
        (
            OVERWRITE_ELEMENT(schema).key("taggedProp")
            .setNewTags((1, 2))
            .commit()
        )


def test_choice_element():
    s = Schema()
    TestStruct1.expectedParameters(s)
    # ------ Optional, defaultValueFromString
    assert s.has("testChoice1") is True
    assert s.isAssignmentOptional("testChoice1") is True
    assert s.hasDefaultValue("testChoice1") is True
    assert s.getDefaultValue("testChoice1") == "RectangleX"
    assert s.isAccessReconfigurable("testChoice1") is True

    assert s.has("CircleX") is True
    assert s.has("CircleX.radius") is True
    assert s.has("RectangleX") is True
    assert s.has("RectangleX.a") is True
    assert s.has("RectangleX.b") is True

    # ------ Optional, noDefaultValue
    assert s.has("testChoice2") is True
    assert s.isAssignmentOptional("testChoice2") is True
    assert s.hasDefaultValue("testChoice2") is False
    assert s.isAccessReconfigurable("testChoice2") is True
    # ------ Mandatory
    assert s.has("testChoice3") is True
    assert s.isAssignmentOptional("testChoice3") is False
    assert s.isAssignmentMandatory("testChoice3") is True
    assert s.hasDefaultValue("testChoice3") is False
    assert s.isAccessReconfigurable("testChoice3") is False
    assert s.isAccessInitOnly("testChoice3") is True
    # ------ Optional, defaultValue
    assert s.has("testChoice4") is True
    assert s.isAssignmentOptional("testChoice4") is True
    assert s.hasDefaultValue("testChoice4") is True
    assert s.getDefaultValue("testChoice4") == "CircleX"
    assert s.isAccessReconfigurable("testChoice4") is True


def test_table_elements():
    sch = Schema()
    OtherSchemaElementsX.expectedParameters(sch)
    assert sch.has("testTable") is True
    assert sch.getMinSize("testTable") == 2
    assert sch.getMaxSize("testTable") == 10
    assert sch.getAliasFromKey("testTable") == [1, 2]
    assert sch.getTags("testTable") == ['hard', 'soft']
    assert sch.getDescription("testTable") == "Test table description"
    assert sch.getDisplayedName("testTable") == 'Test table'
    assert sch.isAssignmentOptional("testTable") is True
    assert sch.isAssignmentMandatory("testTable") is False
    assert sch.hasDefaultValue("testTable") is True
    assert sch.isAccessInitOnly("testTable") is True
    assert sch.getRequiredAccessLevel("testTable") == AccessLevel.EXPERT

    assert sch.has("testTableEmptyDefault") is True
    assert sch.getAliasFromKey("testTableEmptyDefault") == 3
    assert sch.getTags("testTableEmptyDefault") == ['soft']
    assert sch.isAssignmentOptional("testTableEmptyDefault") is True
    assert sch.hasDefaultValue("testTableEmptyDefault") is True
    assert sch.isAccessInitOnly("testTableEmptyDefault") is True
    assert (sch.getRequiredAccessLevel("testTableEmptyDefault") ==
            AccessLevel.EXPERT)

    assert sch.has("testTableNoDefault") is True
    assert (sch.getAliasFromKey("testTableNoDefault") ==
            "aliasForTestTableNoDefault")
    assert sch.getTags("testTableNoDefault") == ['hard']
    assert sch.isAccessReconfigurable("testTableNoDefault") is False
    assert sch.isAccessInitOnly("testTableNoDefault") is True
    assert sch.isAssignmentOptional("testTableNoDefault") is True
    assert sch.hasDefaultValue("testTableNoDefault") is False
    assert (sch.getRequiredAccessLevel("testTableNoDefault") ==
            AccessLevel.OBSERVER)

    assert sch.has("tableMandatory") is True
    assert sch.getAliasFromKey("tableMandatory") == "aliasForTableMandatory"
    assert sch.getTags("tableMandatory") == ['hard']
    assert sch.isAssignmentInternal("tableMandatory") is False
    assert sch.isAssignmentMandatory("tableMandatory") is True
    assert sch.isAccessInitOnly("tableMandatory") is True
    assert (sch.getRequiredAccessLevel("tableMandatory") ==
            AccessLevel.OPERATOR)

    assert sch.has("tableInternal") is True
    assert sch.getAliasFromKey("tableInternal") == "aliasForTableInternal"
    assert sch.getTags("tableInternal") == ['hard']
    assert sch.isAssignmentInternal("tableInternal") is True
    assert sch.isAssignmentMandatory("tableInternal") is False
    assert sch.isAccessReconfigurable("tableInternal") is True
    assert sch.hasDefaultValue("tableInternal") is False
    assert sch.getRequiredAccessLevel("tableInternal") == AccessLevel.OPERATOR

    assert sch.has("tableReadOnly") is True
    assert sch.getAliasFromKey("tableReadOnly") == "aliasForTableReadOnly"
    assert sch.getTags("tableReadOnly") == ['exp', 'hard', 'soft']
    assert sch.isAssignmentOptional("tableReadOnly") is True
    assert sch.isAccessReadOnly("tableReadOnly") is True
    assert sch.hasDefaultValue("tableReadOnly") is True
    assert sch.hasAllowedStates("tableReadOnly") is True
    assert (sch.getAllowedStates("tableReadOnly") ==
            [State.STARTED, State.STOPPED])
    assert sch.getRequiredAccessLevel("tableReadOnly") == AccessLevel.OPERATOR
    assert sch.hasArchivePolicy("tableReadOnly") is True
    assert sch.getArchivePolicy("tableReadOnly") == ArchivePolicy.EVERY_EVENT


def test_imagedata_element():
    sch = Schema()
    SomeClass.expectedParameters(sch)
    h = sch.getParameterHash()
    assert sch.has("imageSetDims") is True
    assert sch.getDisplayedName("imageSetDims") == "Image setDimensions"
    assert (sch.getDescription("imageSetDims") ==
            "ImageData element: setDimensions")
    assert h.getAttribute('imageSetDims.dims', 'defaultValue') == [110, 210]
    assert h.getAttribute('imageSetDims.encoding',
                          'defaultValue') == Encoding.GRAY.value
    assert sch.getRequiredAccessLevel("imageSetDims") == AccessLevel.OPERATOR

    assert sch.has("imageSetType") is True
    assert sch.getRequiredAccessLevel("imageSetType") == AccessLevel.OBSERVER
    assert h.getAttribute('imageSetType.dims', 'defaultValue') == [110, 210]
    assert h.getAttribute('imageSetType.encoding',
                          'defaultValue') == Encoding.GRAY.value

    assert sch.has("imageSetEncoding") is True
    assert sch.getRequiredAccessLevel("imageSetEncoding") == \
        AccessLevel.OPERATOR
    assert h.getAttribute('imageSetEncoding.dims',
                          'defaultValue') == [300, 200]
    assert h.getAttribute('imageSetEncoding.encoding',
                          'defaultValue') == Encoding.RGBA.value

    assert sch.has("imageSetGeometry") is True
    assert sch.getRequiredAccessLevel("imageSetGeometry") == \
        AccessLevel.EXPERT

    assert sch.has("myImageElement") is True
    assert h.getAttribute("myImageElement.dims",
                          'defaultValue') == [500, 800]
    assert sch.getRequiredAccessLevel("myImageElement") == AccessLevel.EXPERT


def test_slot_element():
    sch = Schema()
    TestStruct1.expectedParameters(sch)
    assert sch.has("slotTest") is True
    assert sch.getAliasFromKey("slotTest") == 0x12345678
    assert sch.isCommand("slotTest") is True
    assert sch.getDisplayedName("slotTest") == "Reset"
    assert sch.getTags("slotTest") == ['hard', 'soft']
    assert sch.getDescription("slotTest") == "Test slot element"
    assert sch.getAllowedStates("slotTest") == [
        State.STARTED, State.STOPPED, State.NORMAL]
    assert sch.getRequiredAccessLevel("slotTest") == AccessLevel.OPERATOR
    assert sch.has("slotClean") is True
    assert sch.getAliasFromKey("slotClean") == "aliasClean"
    assert sch.isCommand("slotClean") is True
    assert sch.getDisplayedName("slotClean") == "Clean"
    assert sch.getRequiredAccessLevel("slotClean") == AccessLevel.OBSERVER
    assert sch.has("slotStart") is True
    assert sch.getAliasFromKey("slotStart") == "aliasStart"
    assert sch.getTags("slotStart") == ['hard', 'soft']
    assert sch.isCommand("slotStart") is True
    assert sch.getDisplayedName("slotStart") == "Start"
    assert sch.getRequiredAccessLevel("slotStart") == AccessLevel.EXPERT
    assert sch.has("slotStop") is True
    assert sch.getAliasFromKey("slotStop") == "aliasStop"
    assert sch.isCommand("slotStop") is True
    assert sch.getDisplayedName("slotStop") == "Stop"
    assert sch.getRequiredAccessLevel("slotStop") == AccessLevel.EXPERT


def test_cpp_ndarray():
    # Build numpy arrays on C++ side and convert them to the python objects
    # using commented binding code below  (see "karabind_test.cc")
    # Conversions happen at the last step...

    # m.def("cppNDArray", []() {
    #     const Dims shape(3, 4);
    #     std::vector<int> someData(3 * 4, 7);
    #     for (int i = 0; i < 3; ++i) someData[i] = 100 + i;
    #     NDArray nda(someData.begin(), someData.end(), shape);
    #     return karabind::wrapper::castNDArrayToPy(nda);
    # });
    # m.def("cppNDArrayCopy", []() {
    #     const Dims shape(3, 4);
    #    std::vector<int> someData(3 * 4, 7);
    #    for (int i = 0; i < 3; ++i) someData[i] = 100 + i;
    #    NDArray nda(someData.begin(), someData.end(), shape);
    #    return karabind::wrapper::copyNDArrayToPy(nda);
    # });

    # create py::array (numpy.array) objects
    a = cppNDArray()
    b = cppNDArrayCopy()
    # These binding return the same data
    assert np.all(a == b)
    # The only difference between them is data ownership ...
    # C++ is an owner of data array
    assert a.flags.owndata is False
    assert a.base is not None
    # Python is an owner of data array
    assert b.flags.owndata is True
    assert b.base is None
    assert np.all(a == b)
    # this works like 'deepcopy'
    c = a.copy()
    assert c.flags.owndata is True
    assert c.base is None
    # or use deepcopy directly ..
    d = copy.deepcopy(a)
    assert d.flags.owndata is True
    assert d.base is None
    # and copy works like deepcopy
    e = copy.copy(a)
    assert e.flags.owndata is True
    assert e.base is None



#!/usr/bin/env python
# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# -*- coding: utf-8 -*-

__author__ = "Sergey Esenov <serguei.essenov at xfel.eu>"
__date__ = "$Jul 26, 2012 10:06:25 AM$"

import copy
import os
import signal
import socket
import sys
import threading
import time
import traceback
from itertools import chain
from subprocess import Popen, TimeoutExpired

from karabind import (
    INT32_ELEMENT, NODE_ELEMENT, OVERWRITE_ELEMENT, STRING_ELEMENT,
    VECTOR_STRING_ELEMENT, Broker, EventLoop, Hash, Logger, Schema,
    SignalSlotable, Unit, Validator, generateAutoStartHash, jsonToHash,
    saveToFile)
from karabo.common.api import KARABO_LOGGER_CONTENT_DEFAULT, ServerFlags

from .configurator import Configurator
from .decorators import KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS
from .device import PythonDevice
from .plugin_loader import DEFAULT_NAMESPACE, PluginLoader
from .runner import Runner


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("DeviceServer", "1.0")
class DeviceServer:
    """Device server serves as a launcher of python devices.

    It scans 'plugins' directory for new plugins (python scripts) available and
    communicates its findings to master device server.  It communicates XSD
    form of schema of user devices and starts such devices as separate process
    if user push "Initiate" button in GUI
    """
    instanceCountLock = threading.Lock()
    instanceCountPerDeviceServer = dict()

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("serverId")
            .displayedName("Server ID")
            .description("The device-server instance id uniquely identifies a"
                         " device-server instance in the distributed system")
            .assignmentOptional().noDefaultValue()
            .commit(),

            STRING_ELEMENT(expected).key("hostName")
            .displayedName("Forced Hostname")
            .description(
                "The hostname can be optionally forced to a specific string. "
                "The host's definition will be used if not specified.")
            .assignmentOptional().noDefaultValue()
            .expertAccess()
            .init()
            .commit(),

            INT32_ELEMENT(expected).key("heartbeatInterval")
            .displayedName("Heartbeat interval")
            .description("The heartbeat interval")
            .assignmentOptional().defaultValue(20)
            .minInc(10)  # avoid too much traffic
            .expertAccess()
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("deviceClasses")
            .displayedName("Device Classes")
            .description("The device classes the server will manage")
            .assignmentOptional()
            .defaultValue([])
            .expertAccess()
            .commit(),

            STRING_ELEMENT(expected).key("init")
            .displayedName("Auto start")
            .description("Auto starts selected devices")
            .assignmentOptional().defaultValue("")
            .commit(),

            STRING_ELEMENT(expected).key("pluginNamespace")
            .displayedName("Plugin Namespace")
            .description("Namespace to search for plugins")
            .assignmentOptional().defaultValue(DEFAULT_NAMESPACE)
            .expertAccess()
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("serverFlags")
            .displayedName("Server Flags")
            .description("ServerFlags describing the device server, "
                         "the values must correspond to the enum ServerFlags")
            .assignmentOptional()
            .defaultValue([])
            .expertAccess()
            .commit(),

            NODE_ELEMENT(expected).key("log")
            .description("Logging settings")
            .displayedName("Logger")
            .appendParametersOf(Logger)
            .commit(),

            OVERWRITE_ELEMENT(expected).key("log.file.filename")
            # Will be assembled programmatically from environment and serverId
            .setNewAssignmentInternal()
            .commit(),

            STRING_ELEMENT(expected).key("timeServerId")
            .description("The instance id uniquely identifies a TimeServer"
                         " instance in the distributed system")
            .displayedName("TimeServer ID")
            .assignmentOptional().defaultValue("")
            .commit(),

            INT32_ELEMENT(expected).key("instantiationTimeout")
            .displayedName("Instantiation Timeout")
            .description("How long to wait for device coming up before "
                         "slotStartDevice fails")
            .unit(Unit.SECOND)
            .assignmentOptional().defaultValue(10)
            .commit(),
        )

    def signal_handler(self, signum, frame):
        if signum == signal.SIGINT:
            print('INTERRUPT : You pressed Ctrl-C!')
        else:
            print('INTERRUPT : You terminated me!')
        if self._sigslot is not None:
            # Better do not go via self._sigslot.call("", "slotKillServer"),
            # otherwise it will run later in another thread.
            self.slotKillServer()
        else:
            self.stopDeviceServer()

    def __init__(self, config):
        """Constructor"""
        if config is None:
            raise ValueError(
                "Input configuration for constructor should be Hash, not None")
        super().__init__()
        self._sigslot = self.logger = None
        self.availableDevices = dict()
        self.deviceInstanceMap = dict()
        self._startingDevices = dict()  # needs protection by lock below
        self.deviceInstanceMapLock = threading.RLock()
        self.instantiationTimeout = config["instantiationTimeout"]
        if config.get('hostName') is not None:
            self.hostname = config['hostName']
        else:
            self.hostname = socket.gethostname().partition('.')[0]
        # Check if obsolete syntax still used...
        if config.has('autoStart'):
            raise KeyError(
                "'autoStart' syntax not supported anymore, use 'init'")
        self.autoStart = None
        if config.get("init") != "":
            cfg = jsonToHash(config.get("init"))
            asv = generateAutoStartHash(cfg)
            self.autoStart = asv['autoStart']

        self.deviceClasses = config.get("deviceClasses")
        self.pluginNamespace = config.get("pluginNamespace")
        self.timeServerId = config.get("timeServerId")

        if 'serverId' in config:
            self.serverId = config['serverId']
        else:
            self.serverId = self._generateDefaultServerId()

        self.connectionParameters = Hash(Broker.brokerTypeFromEnv(),
                                         Hash("instanceId", self.serverId))
        self.loggerParameters = None  # assemble in loadLogger
        self.pid = os.getpid()
        self.seqnum = 0

        # Start the logging system before the scanning of plugins, so any
        # error during plugin loading can be logged.
        self.loadLogger(config)
        self.logger = Logger.getLogger(self.serverId)

        info = Hash("type", "server")
        info["serverId"] = self.serverId
        info["version"] = self.__class__.__version__
        info["host"] = self.hostname
        info["lang"] = "bound"
        info["log"] = config.get("log.level")

        self.serverFlags = config.get("serverFlags")
        serverFlags = 0
        for flag in self.serverFlags:
            if flag in ServerFlags.__members__:
                serverFlags |= ServerFlags[flag]
            else:
                raise NotImplementedError(
                    f"Provided serverFlag is not supported: {flag}")
        info["serverFlags"] = serverFlags

        devicesInfo, scanLogs = self.scanPlugins(self.pluginNamespace)
        info.merge(devicesInfo)

        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

        # Be aware: If signal handling is triggered before __init__ has
        # finished, self._sigslot might be set to None, causing problems below,
        # e.g. "'NoneType' object has no attribute 'getConnection'".
        # In worst case, SignalSlotable is constructed, but not yet assigned
        # to self._sigslot when the signal handler wants to reset it to None.
        # But the process stops nevertheless due to the EventLoop.stop().

        self._sigslot = SignalSlotable(
            self.serverId, self.connectionParameters,
            config["heartbeatInterval"], info)

        # Register before sigslot.start(), i.e before sending instanceNew:
        self._registerAndConnectSignalsAndSlots()

        # Start SignalSlotable object - multithreading begins
        self._sigslot.start()

        # Now we can log the postponed logging messages - could have been
        # done directly after assigning self.logger, but prefer to do after
        # instanceNew triggered by self._sigslot.start():
        for level, message in scanLogs:
            getattr(self.logger, level)(message)

        msg = "DeviceServer starts on host '{0.hostname}' " \
              "with pid {0.pid}, broker: {1}"
        self.logger.info(msg.format(
            self, self._sigslot.getConnection().getBrokerUrl()))

        self.doAutoStart()

    def _generateDefaultServerId(self):
        return self.hostname + "_Server_" + str(os.getpid())

    def loadLogger(self, inputCfg):
        self.loggerParameters = copy.copy(inputCfg["log"])
        # The file logger filename is completely specified here.
        path = os.path.join(os.environ['KARABO'], "var", "log", self.serverId)
        if not os.path.isdir(path):
            os.makedirs(path)
        path = os.path.join(path, 'device-server.log')
        self.loggerParameters.set('file.filename', path)

        Logger.configure(self.loggerParameters)
        Logger.useConsole()
        Logger.useFile()
        Logger.useCache()

    def _registerAndConnectSignalsAndSlots(self):
        self._sigslot.registerSlot(self.slotStartDevice)
        self._sigslot.registerSlot(self.slotKillServer)
        self._sigslot.registerSlot(self.slotDeviceUp)
        self._sigslot.registerSlot(self.slotDeviceGone)
        self._sigslot.registerSlot(self.slotGetClassSchema)
        self._sigslot.registerSlot(self.slotLoggerLevel)
        self._sigslot.registerSlot(self.slotLoggerContent)

    def scanPlugins(self, pluginNamespace):
        """Scan for available device classes

        Returns Hash with keys "deviceClasses" and "visibilities" to be merged
        into instance info and a list of log messages to be send.
        Inside the list there are tuples of two string: log level (e.g. "info")
        and message.
        Also fills self.availableDevices dictionary.
        """
        loaderCfg = Hash("pluginNamespace", pluginNamespace)
        loader = PluginLoader.create("PythonPluginLoader", loaderCfg)
        entrypoints = loader.update()
        logs = [("error",
                 f"scanPlugins: Cannot load device plugin {name} -- {exc}")
                for name, exc in loader.plugin_errors.items()]
        for ep in entrypoints:
            if not self.deviceClasses or ep.name in self.deviceClasses:
                # All entry points have been verified already
                deviceClass = ep.load()
                classid = deviceClass.__classid__
                try:
                    schema = Configurator(PythonDevice).getSchema(classid)
                    self.availableDevices[classid] = {"module": ep.name,
                                                      "schema": schema}
                    logs.append(("info",
                                 'Successfully loaded plugin: "{}"'.format(
                                     ep.name)))
                # A generic handler is used here due to the different kind
                # of exceptions that may be raised when obtained the schema
                # for the just loaded plugin.
                except Exception as e:
                    m = "Failure while building schema for class {}, base " \
                        "class {} and bases {} : {}" \
                        .format(classid, deviceClass.__base_classid__,
                                deviceClass.__bases_classid__, repr(e))
                    # repr(e) also includes type
                    logs.append(("error", m))

        instInfo = Hash("deviceClasses",
                        [classid for classid in self.availableDevices.keys()])

        return instInfo, logs

    def doAutoStart(self):
        if self.autoStart is None:
            return
        for entry in self.autoStart:
            # entry is a Hash with a single key which is the classid
            classId = entry.getKeys()[0]
            if classId in self.availableDevices:
                config = entry.get(classId)
                hsh = Hash("classId", classId,
                           "configuration", config)
                if 'deviceId' in config:
                    hsh.set('deviceId', config.get('deviceId'))
                self.instantiateDevice(hsh)

    def stopDeviceServer(self):
        # HACK: Sometimes the C++ destructor of SignalSlotable is not called
        #       and then instanceGone is not sent. This can happen when many
        #       variables are still referring to the SignalSlotable here.
        #       This has been seen when stopped early in the initialisation
        #       procedure, leading to a very high (>27000) refcount.
        #       If we are the last ones, it is guaranteed that the destructor
        #       is called.
        #
        # Take over SignalSlotable to avoid new references in other threads:
        localSigSlot = self._sigslot
        self._sigslot = None
        refCount = sys.getrefcount(localSigSlot)
        if refCount > 2:
            # count of 2: localSigSlot and the one internal to getrefcount
            print("Forced call to slotInstanceGone due to refcount", refCount)
            localSigSlot.call("*", "slotInstanceGone",
                              self.serverId, localSigSlot.getInstanceInfo())
        # HACK end

        EventLoop.stop()

    def slotStartDevice(self, configuration):
        self.instantiateDevice(configuration)

    def instantiateDevice(self, input_config):
        classId = input_config['classId']

        # Get configuration
        config = copy.copy(input_config['configuration'])

        # Inject serverId
        config['serverId'] = self.serverId

        # Inject deviceId
        config['deviceId'] = input_config.get("deviceId", default="")
        if not config['deviceId']:
            config.set('deviceId',
                       self._generateDefaultDeviceInstanceId(classId))

        self.logger.info("Trying to start a '{}' with device id '{}'"
                         "...".format(classId, config['deviceId']))
        self.logger.debug(
            f"with the following configuration:\n{input_config}")

        # Inject HostName
        config['hostName'] = self.hostname

        # If not explicitely specified, let device inherit logger level
        if not config.has("log.level"):
            config["log.level"] = self.loggerParameters["level"]

        # Before starting device process, validate config
        schema = Configurator(PythonDevice).getSchema(classId)
        validator = Validator()
        (ok, msg, _) = validator.validate(schema, config)
        if not ok:
            msg = msg.strip()  # cut-off trailing newline...
            self.logger.warning(
                f"Failed to start '{config['deviceId']}': {msg}")
            self._sigslot.reply(ok, msg)
            return

        # Now few config manipulation intentionally AFTER 'validate':
        # Add connection type and parameters used by device server for
        # connecting to broker.
        config['_connection_'] = self.connectionParameters

        # Add temporary namespace variable
        config['_pluginNamespace_'] = self.pluginNamespace

        # Add time server ID configured for device server.
        config['timeServerId'] = self.timeServerId

        # Also add config for Logger appenders
        for appender in ["console", "file", "cache"]:
            config["_logger_." + appender] = self.loggerParameters[appender]

        reply = self._sigslot.createAsyncReply()

        try:
            self._launchDevice(config, classId, reply)
        except Exception as e:
            msg = f"Device of class '{classId}' could not be started: {e}"
            details = traceback.format_exc()
            self.logger.warning(f"{msg}:\nFailure details:\n{details}")
            # Cannot call AsyncReply object directly in slot, so post:
            EventLoop.post(lambda: reply.error(msg, details))

    def _launchDevice(self, config, classId, reply):

        deviceid = config["deviceId"]  # exists, see instantiateDevice

        modname = self.availableDevices[classId]["module"]

        # Create unique filename in /tmp - without '/' from deviceid...
        # .bin indicates binary format:
        # XML has trouble with VECTOR_STRING where string contains comma
        while True:
            filename = (f"/tmp/{modname}.{classId}."
                        f"{deviceid.replace(os.path.sep, '_')}"
                        f".configuration_{self.pid}_{self.seqnum}.bin")
            if os.path.isfile(filename):
                self.seqnum += 1
            else:
                break

        saveToFile(config, filename)
        params = [modname, classId, filename]

        with self.deviceInstanceMapLock:
            if deviceid in self.deviceInstanceMap:
                # Already there! Check whether previous process is alive:
                prevLauncher = self.deviceInstanceMap[deviceid]
                if prevLauncher.child.poll() is None:
                    # Process still up. Check Karabo communication by ping:
                    request = self._sigslot.request(deviceid, "slotPing", 1)
                    try:
                        # Too lazy to use async techniques for this corner case
                        request.waitForReply(3000)  # in milliseconds
                    except TimeoutError:
                        # Indeed dead Karabo-wise:
                        self.logger.warning("Kill previous incarnation of "
                                            f"'{deviceid}' since reply to "
                                            "ping timed out.")
                        prevLauncher.kill()
                    except Exception as e:
                        # Unexpected exception - give up
                        raise e
                    else:
                        # Technically, it could be alive on another server,
                        # but who cares about that detail...
                        # Just raising RuntimeError is in principle also OK
                        # Cannot call AsyncReply() directly in slot, so post:
                        EventLoop.post(lambda: reply(False,
                                                     f"{deviceid} already "
                                                     "instantiated and alive"))
                        return

            launcher = Launcher(params)
            launcher.start()
            self.deviceInstanceMap[deviceid] = launcher
            # Keep track of being starting to reply when succeeded (or timeout)
            self._startingDevices[deviceid] = (reply, time.time())
            if len(self._startingDevices) == 1:  # otherwise already running
                EventLoop.post(self._checkStartingDevices, 1)

    def _checkStartingDevices(self):
        failed = []
        somethingLeft = False
        with self.deviceInstanceMapLock:
            for devId, (reply, startedAt) in self._startingDevices.items():
                if (time.time() - startedAt) > self.instantiationTimeout:
                    failed.append((devId, reply))
            for devId, _ in failed:
                del self._startingDevices[devId]
                # But don't touch self.deviceInstanceMap - device may be "late"
            if self._startingDevices:
                somethingLeft = True

        # Reply without lock since sending messages is blocking:
        for devId, reply in failed:
            msg = (f"Timeout of instantiation: {devId} did not confirm it is "
                   f"up within {self.instantiationTimeout} seconds")
            self.logger.warning(msg)
            reply(False, msg)

        if somethingLeft:
            EventLoop.post(self._checkStartingDevices, 0.5)

    def slotKillServer(self):
        if self.logger:
            self.logger.info("Received kill signal")
        else:  # might get killed by signal handler before setting up logging
            print("Received kill signal")
        with self.deviceInstanceMapLock:
            # Break "loop" to check for starting devices
            self._startingDevices.clear()
            # Loop twice: First to quickly tell all devices to go down and then
            #             to wait until they are indeed down (or need killing)
            for deviceid in self.deviceInstanceMap:
                self._sigslot.call(deviceid, "slotKillDevice")
            for deviceid, launcher in self.deviceInstanceMap.items():
                if launcher:
                    try:
                        launcher.join()
                    except TimeoutExpired:
                        self.logger.warning("Timeout on server shutdown while"
                                            " stopping the process for '{}'"
                                            "... SIGKILL".format(deviceid))
                        launcher.kill()
            self.deviceInstanceMap = {}
        try:
            self._sigslot.reply(self.serverId)
        except Exception as e:
            if self.logger:  # see above
                msg = ("Did not notify distributed system of server shutdown:"
                       "\n {}").format(e)
                self.logger.error(msg)
        finally:
            self.stopDeviceServer()

    def slotDeviceUp(self, instanceId, success, reason):
        "Slot for device to tell us whether it managed to get alive"
        awaited = None
        with self.deviceInstanceMapLock:
            awaited = self._startingDevices.pop(instanceId, None)
        if awaited:
            asyncReply = awaited[0]
            if success:
                asyncReply(success, instanceId)
            else:
                msg = (f"could not instantiate device {instanceId}."
                       f" Reason: {reason}")
                asyncReply(success, msg)
        else:
            self.logger.warning(f"Unexpected slotDeviceUp for {instanceId}")
            # No need to inform caller about failure via raising an exception

    def slotDeviceGone(self, instanceId):
        self.logger.info(
            f"Device '{instanceId}' notifies '{self.serverId}' about "
            "its future death.")
        with self.deviceInstanceMapLock:
            if instanceId in self.deviceInstanceMap:
                launcher = self.deviceInstanceMap[instanceId]
                if launcher:
                    try:
                        launcher.join()
                    except TimeoutExpired:
                        self.logger.warning("Timeout on device shutdown while"
                                            " stopping the process for '{}'"
                                            "... SIGKILL".format(instanceId))
                        launcher.kill()
                del self.deviceInstanceMap[instanceId]
            self.logger.info("Device '{}' removed from server."
                             .format(instanceId))

    def slotGetClassSchema(self, classId):
        try:
            schema = Configurator(PythonDevice).getSchema(classId)
            self._sigslot.reply(schema, classId, self.serverId)
        except AttributeError as e:
            self.logger.warning(f"Replied empty schema due to : {str(e)}")
            self._sigslot.reply(Schema(), classId, self.serverId)

    def slotLoggerLevel(self, newlevel):
        # In contrast to C++, the new level will not be "forwarded" to
        # existing devices. Python devices have their own slotLoggerLevel
        # which allows level setting device by device.
        # But future device instantiations should inherit their level from
        # the current value of the server:
        oldprio = Logger.getLevel()
        Logger.setLevel(newlevel)
        self.logger.info(
            f"log Level changed : {oldprio} ==> {newlevel}")
        # Also devices started in future get the new level by default:
        self.loggerParameters["level"] = newlevel
        # Merge the new log level into the instanceInfo
        self._sigslot.updateInstanceInfo(Hash("log", newlevel))

    def slotLoggerContent(self, info):
        """Slot call to receive logger content from the CacheLogger

        replies with a Hash containing a key, `serverId`
        and a `content` containing a vector of hashes formatted in the
        same way the broker based logging uses. For details:
        ``src/karabo/log/CacheLogger.cc``

        :param info: input Hash containing an optional `logs` integer
                     defining the maximum number of lines returned
        """
        # create a map where to store the devices' logs
        with self.deviceInstanceMapLock:
            log_map = dict.fromkeys(self.deviceInstanceMap)

        if not log_map:  # no devices => direct reply instead of an async one
            nMessages = info.get(
                "logs", default=KARABO_LOGGER_CONTENT_DEFAULT)
            content = Logger.getCachedContent(nMessages)
            self._sigslot.reply(
                Hash("serverId", self.serverId, "content", content))
            return

        replyLock = threading.Lock()

        areply = self._sigslot.createAsyncReply()
        # copy because the info object will be cleaned up after
        # this function has returned
        info = copy.copy(info)

        class Handler:
            def __init__(s, deviceId):
                s.deviceId = deviceId

            def on_reply(s, reply):
                # copy because the reply object will be
                # cleaned up after the handler has returned
                _copy = [copy.copy(entry) for entry in reply['content']]
                log_map[s.deviceId] = _copy
                s.done()

            def on_error(s, msg, details):
                msg = (f"Missing Logger Content from '{s.deviceId}' : "
                       f"{msg} - DETAILS {details}")
                self.logger.warning(msg)
                log_map.pop(s.deviceId, None)
                s.done()

            def done(s):
                # make sure that only one handler will call
                # `_replySlotLoggerContent`.
                with replyLock:
                    if all(log_map.values()):
                        self._replySlotLoggerContent(info, log_map, areply)

        for deviceId in log_map.keys():
            # request the content from the devices in the server.
            req = self._sigslot.request(deviceId, "slotLoggerContent", info)
            h = Handler(deviceId)
            # timeout smaller than GUI one,
            # otherwise that misses all logs and not only those of bad device
            req.receiveAsync(h.on_reply, h.on_error, timeoutMs=4500)

    def _replySlotLoggerContent(self, info, log_map, areply):
        nMessages = info.get("logs", default=KARABO_LOGGER_CONTENT_DEFAULT)
        content = list(log_map.values())
        # add the latest server logs.
        content.append(Logger.getCachedContent(nMessages))
        content = chain.from_iterable(content)
        # sort by timestamp and get the last "nMessages" entries
        content = sorted(content, key=lambda entry: entry['timestamp'])
        # reply
        areply(Hash("serverId", self.serverId,
                    "content", content[-1 * nMessages:]))

    def _generateDefaultDeviceInstanceId(self, devClassId: str) -> str:
        cls = self.__class__
        with cls.instanceCountLock:
            count = cls.instanceCountPerDeviceServer.setdefault(
                self.serverId, 0) + 1
            cls.instanceCountPerDeviceServer[self.serverId] = count
            tokens = self.serverId.split("_")
            if tokens[-1] == str(os.getpid()):
                instance_id = f"{tokens[0]}-{tokens[-2]}_{devClassId}_{count}"
            else:
                instance_id = f"{self.serverId}_{devClassId}_{count}"

            return instance_id


class Launcher:
    """ Launches a PythonDevice in its own interpreter."""

    def __init__(self, params):
        self.args = params

    def start(self):
        args = [sys.executable, "-c"]
        args.append("from karabo.bound.device import launchPythonDevice;"
                    "launchPythonDevice()")
        args.extend(self.args)

        self.child = Popen(args)

    # TODO This should be renamed, has nothing to do with threads
    def join(self):
        if self.child.poll() is None:
            self.child.wait(timeout=5)

    def kill(self):
        self.child.kill()
        self.child.wait()


def main(argv=None):
    try:
        # Change directory to $KARABO/var/data
        os.chdir(os.path.join(os.environ['KARABO'], 'var', 'data'))
    except KeyError:
        print("ERROR: $KARABO is not defined. Make sure you have sourced the "
              "'activate' script.")
        return 1

    argv = argv or sys.argv
    # Load plugins already here to make them available for -h option
    PluginLoader.create("PythonPluginLoader", Hash()).update()
    t = threading.Thread(target=EventLoop.work)
    t.start()
    server = None
    try:
        server = Runner(DeviceServer).instantiate(argv)
        if not server:
            EventLoop.stop()
            # We get here with -h option: Avoid print after finally.
            return 0
    except Exception:
        EventLoop.stop()
        traceback.print_exc()
        return 2
    finally:
        del server  # increase chance that no log appears after print below
        t.join()
    print(os.path.basename(argv[0]), "has exited!\n")
    return 0


if __name__ == '__main__':
    sys.exit(main())



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from .configurator import Configurator


def KARABO_CLASSINFO(classid, version):
    """
    This decorator should be placed just before class definition. It adds some
    variables used in KARABO configuration infrastructure. It has two
    parameters: "classId" and "version" similar to corresponding C++ macro:
    Example:
            @KARABO_CLASSINFO("Shape","1.0")
            class Shape(object):
                ...
    """
    def realDecorator(theClass):
        if isinstance(theClass, type):
            theClass.__classid__ = str(classid)
            theClass.__version__ = version
            if hasattr(theClass, "__base_classid__"):
                Configurator(theClass.__base_classid__).registerClass(theClass)
                theClass.__bases_classid__ = []
                for base in theClass.__bases__:
                    for i in getattr(base, '__bases_classid__', ()):
                        if i not in theClass.__bases_classid__:
                            theClass.__bases_classid__.append(i)
                theClass.__bases_classid__ = tuple(theClass.__bases_classid__)
        return theClass
    return realDecorator


def KARABO_CONFIGURATION_BASE_CLASS(theClass):
    """
    This decorator should be placed just before "KARABO_CLASSINFO" decorator.
    It registers the class as the base configurable class and adds the
    following classmethods: "create", "createNode", "getSchema" and
    "getRegisteredClasses". It has no parameters.
    Example:
            @KARABO_CONFIGURATION_BASE_CLASS
            @KARABO_CLASSINFO("Shape","1.0")
            class Shape(object):
                ...
    """
    if isinstance(theClass, type):
        Configurator.registerAsBaseClass(theClass)

        def create(cls, *args):
            """
            The factory classmethod to create the instance of class with
            "classId" using input "configuration".
            Example:
                instance = Shape.create("EditableCircle",
                                        Hash("radius", 12.345))

            The factory classmethod to create instance according input
            "configuration".  The configuration should have "classId"
            of class to be created as a root element.
            Example:
                    configuration = Hash("EditableCircle.radius", 12.345)
                    instance = Shape.create(configuration)
            """
            if len(args) == 1:
                return Configurator(cls.__base_classid__).create(args[0])
            elif len(args) == 2:
                return Configurator(cls.__base_classid__).create(args[0],
                                                                 args[1])
            elif len(args) == 3:
                return Configurator(cls.__base_classid__).create(args[0],
                                                                 args[1],
                                                                 args[2])
            else:
                raise TypeError("Wrong number of arguments or their types")

        create = classmethod(create)
        theClass.create = create

        def createNode(cls, nodename, classid, configuration):
            """
            The helper classmethod to create instance of class specified by
            "classId" using sub-configuration specified by "nodeName" which has
            to be a part of input "configuration".
            """
            return Configurator(cls.__base_classid__).createNode(nodename,
                                                                 classid,
                                                                 configuration)

        createNode = classmethod(createNode)
        theClass.createNode = createNode

        def getSchema(cls, classid, rules=None):
            """
            Use this classmethod to get schema for class with "classid" using
            assembly "rules"
            Example:
                    schema = Shape.getSchema("Rectangle")
                or
                    schema = Shape.getSchema("Rectangle", AssemblyRules())
            """
            return Configurator(cls.__base_classid__).getSchema(classid, rules)

        getSchema = classmethod(getSchema)
        theClass.getSchema = getSchema

        def getRegisteredClasses(cls):
            return Configurator(cls.__base_classid__).getRegisteredClasses()

        getRegisteredClasses = classmethod(getRegisteredClasses)
        theClass.getRegisteredClasses = getRegisteredClasses

    return theClass
